//
//  TrackerDocument.m
//  TomatoTorrent
//
//  Created by Sarwat Khan on Tue Jun 10 2003.
//  Copyright (c) 2003, Sarwat Khan. All rights reserved.
//  
//  Redistribution and use in source and binary forms, with or without
//  modification, are permitted provided that the following conditions are
//  met:
//          
//      1. Redistributions of source code must retain the above copyright
//         notice, this list of conditions and the following disclaimer.
//          
//      2. Redistributions in binary form must reproduce the above
//         copyright notice, this list of conditions and the following disclaimer
//         in the documentation and/or other materials provided with the
//         distribution.
//          
//      3. Neither Sarwat Khans name nor the names of any contributors to
//         the Tomato Torrent Software may be used to endorse or promote
//         products derived from this software without specific prior written
//         permission.
//  
//  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
//  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
//  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
//  A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT
//  OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
//  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
//  LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
//  DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
//  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
//  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
//  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


#import "TrackerDocument.h"

#import "TomatoTorrentAgent.h"
#import "TrackerWindowController.h"
#import "TrackerDocumentController.h"
#import "TrackerFaultFormatter.h"
#import "NSDictionary+DataTransform.h"

#define kUserDefault_lastTrackerDocumentURL 	@"LastTrackerDocumentURL"

#define	kLogFileName				@"log.text"
#define kTrackStateFileName			@"track_state"
#define kDefaultsFileName			@"defaults.plist"

@interface TrackerDocument (Private)
- _lastDocumentPath;
- _setLastDocumentPath:path;
@end


@implementation TrackerDocument

- (id)init
{
    if (self = [super init]) 
		_agent = [TomatoTorrentAgent new];
    return self;
}

- (void) dealloc
{
	[_agent shutdown];
	[_agent release];
	[_logPath release];
	[_lastDocumentPath release];
	
	[super dealloc];
}

- setTrackerWindowController:controller
{
	_trackerController = [controller trackerDocumentController];
	[self addWindowController:controller];
	return self;
}

- (void)makeWindowControllers;
{
	id	windowController;
	
	if (_trackerController == nil)
	{
		windowController = [TrackerDocumentController makeTrackerDocumentWindowController];
		[self addWindowController:windowController];
		
		_trackerController = [windowController trackerDocumentController];
		[windowController release];
	}
}

- _fileWrapperDictionary:defaults
{
	id		result;
	id		path, file;

	result = [NSMutableDictionary dictionaryWithObject:defaults forKey:kDefaultsFileName];

	if (path = [self _lastDocumentPath])
	{
		//add other files if they exist.
		if (file = [[NSFileWrapper alloc] initWithPath:[path stringByAppendingPathComponent:kLogFileName]])

			[result setObject:file forKey:kLogFileName],
			[file release];
		
		if (file = [[NSFileWrapper alloc] initWithPath:[path stringByAppendingPathComponent:kTrackStateFileName]])

			[result setObject:file forKey:kTrackStateFileName],
			[file release];
	}
		
	return result;
}

//	the old files move on us while saving to a temporary backup location. for the above
//	method to work, it has to get the files from the backup file wrapper.

- (BOOL)writeToFile:(NSString *)fullDocumentPath 
			 ofType:(NSString *)docType 
	   originalFile:(NSString *)fullOriginalDocumentPath 
	  saveOperation:(NSSaveOperationType)saveOperationType
{
	[self _setLastDocumentPath:fullOriginalDocumentPath];

	return [super writeToFile:fullDocumentPath 
					   ofType:docType 
				 originalFile:fullOriginalDocumentPath 
				saveOperation:saveOperationType];
}


- (NSFileWrapper *)fileWrapperRepresentationOfType:(NSString *)type;
{
	id		result = nil;
	id		docs, 			//main wrapper dictionary
			defaults, 		//settings dictionary
			settings,		//settings
			plistData;		//nsdata of settings
	
	if ([type isEqual:kTrackerDocumentType])
	{
		//	file wrapper contains plist of defaults and the --dfile for the tracker.
		//	we only maintain the defaults; the tracker will write the dfile.

		settings = [_agent valuesForKeys:
			[NSArray arrayWithObjects:@"trackerPort", @"trackerAllowedDirPath", @"trackerOptions", nil]];
		
		plistData	= [(id)CFPropertyListCreateXMLData(NULL, settings) autorelease];
		defaults 	= [[[NSFileWrapper alloc] initRegularFileWithContents:plistData] autorelease];
		docs 		= [self _fileWrapperDictionary:defaults];
		result 		= [[[NSFileWrapper alloc] initDirectoryWithFileWrappers:docs] autorelease];
	}
	
	return result;
}

- (BOOL)loadFileWrapperRepresentation:(NSFileWrapper *)wrapper ofType:(NSString *)type;
{
	BOOL	result = NO;
	id 		defaultsFile, settings;
	
	if ([wrapper isDirectory] && [type isEqual:kTrackerDocumentType])
	{
		result = YES;

		if (defaultsFile = [[wrapper fileWrappers] objectForKey:kDefaultsFileName])
		{
			if ([defaultsFile isRegularFile])
			{
				settings = (id)CFPropertyListCreateFromXMLData(
								NULL,
								(CFDataRef)[defaultsFile regularFileContents],
								kCFPropertyListImmutable,
								NULL);
				if (settings)
				{
					[_agent takeValuesFromDictionary:settings];
				}
			}
		}
	}
	
	return result;
}

+ lastTrackerDocumentURL;
{
	id		data;
	id		result = nil;
	
	if (data = [[NSUserDefaults standardUserDefaults] dataForKey:kUserDefault_lastTrackerDocumentURL])
	{
		result = [NSUnarchiver unarchiveObjectWithData:data];
		
		if ([result isKindOfClass:[NSURL class]] == NO)
			result = nil;
	}
	
	return result;
}

- lastTrackerDocumentURL
{
	return [[self class] lastTrackerDocumentURL];
}

- _setLastTrackerDocumentURL:url
{
	NSData	*data;
	
	if (data = [NSArchiver archivedDataWithRootObject:url])
	{
		[[NSUserDefaults standardUserDefaults] setObject:data
												  forKey:kUserDefault_lastTrackerDocumentURL];
	}
	
	return self;
}

- settings;
{
	return [_agent trackerOptions];
}

id restoreDefaultSettings_map(id object, id key, void *context)
{
	return [object objectAtIndex:kTomatoAgentModuleDefaultsEntryIndex_value];
}
- restoreDefaultSettings;
{

	[_agent setTrackerOptions:[[_agent defaultTrackerOptions] 
		dictionaryMappingObjectsUsingFunction:restoreDefaultSettings_map 
									  context:nil]];
	return self;
}

- startTracker:sender;
{
	id		result = nil;
	id		path,
			log,
			fileManager = [NSFileManager defaultManager];
	BOOL	isDir;
	
	if (path = [self fileName])
	{
		if ([fileManager fileExistsAtPath:path isDirectory:&isDir] 
			&& isDir 
			&& [fileManager isWritableFileAtPath:path])
		{
			FILE		*temp;
			
			log = [path stringByAppendingPathComponent:kLogFileName];
			
			if (temp = fopen([log cString], "a"))
			{
				id		startLog = SKLocalizedString(@"==> Starting Tracker At %@", @"tracker start log message at (date)");
				id		string = [NSString stringWithFormat:startLog, [NSCalendarDate date]];
				
				fprintf(temp, "\n\n%s\n\n", [string cString]);
				fclose(temp);
				
				//set the log and state files
				[_agent setTrackerLogFilePath:log];
				
				//set the state file.
				[_agent setTrackerStateFilePath:[path stringByAppendingPathComponent:kTrackStateFileName]];
				
				if ([_agent startTracker:self])
					result = self;	//else log file entry is misleading, but it shouldn't be backed up anyway.
				
				_logPath = [log retain];
				
				[_trackerController showTrackerProgress:_logPath];
					
				[self _setLastTrackerDocumentURL:[NSURL fileURLWithPath:path]];
			}
		}
	}
	
	return result;
}

- stopTracker:sender;
{
	if ([_agent stopTracker:self])
		return self;

	return nil;
}

- trackerLogFilePath
{
	return _logPath;
}

- setTrackerPort:(int)port;
{
	[_agent setTrackerPort:port];
	return self;
}

- (int)trackerPort;
{
	return [_agent trackerPort];
}

- setTrackerAllowedDirPath:path;
{
	[_agent setTrackerAllowedDirPath:path];
	return self;
}

- trackerAllowedDirPath;
{
	return [_agent trackerAllowedDirPath];
}

- (BOOL) trackerIsRunning;
{
	return [_agent didStartTracker];
}

@end

@implementation TrackerDocument (TomatoTorrentAgentDelegate)

- _writeTrackerStoppedToLogWithTitle:titleString description:descString
{
	if (_logPath)
	{
		FILE	*temp;
		
		if (temp = fopen([_logPath cString], "a"))
		{
			fprintf(temp, "\n%s\n%s\n\n", [titleString cString], [descString cString]);
			fclose(temp);
		}
	}
	
	return self;
}

- tomatoTorrentAgent:agent trackerStartup:response;
{
	id		fault;
	
	if (fault = [response fault])
	{
		id		faultLog = SKLocalizedString(@"==> Tracker Failed to Start", @"tracker failed start log message");
		id		formatter = [[TrackerFaultFormatter new] autorelease];
		id		message = [formatter stringForObjectValue:fault];	
		
		[self _writeTrackerStoppedToLogWithTitle:faultLog description:message];
		[_trackerController stopShowingTrackerProgress];
	}
	
	return self;
}

- tomatoTorrentAgent:agent trackerStopped:response;
{
	if ([response fault] == nil)
	{
		id		title = SKLocalizedString(@"==> Tracker Stopped", @"tracker stopped log message title");
		id		desc = SKLocalizedString(@"Normal shutdown.", @"tracker stopped log message description");
	
		[self _writeTrackerStoppedToLogWithTitle:title description:desc];
		[_trackerController stopShowingTrackerProgress];
		[_logPath release], _logPath = nil;
	}
	else
		NSLog(@"caught fault when trying to stop tracking, \n%@", [response fault]);	//+++ should show user.
		
	return self;
}

- tomatoTorrentAgent:agent trackerState:response;
{
	//if the tracker is not running tell the gui to stop tracking.
	if ([response fault] == nil && _logPath != nil)
	{
		if ([[[response responseContent] objectForKey:kTomatoAgentKey_trackerState_status] 
			isEqual:kTomatoAgentKey_trackerState_status_running] == NO)
		{
			id		title = SKLocalizedString(@"==> Tracker Not Running", @"tracker state not running log message title");
			id		desc = SKLocalizedString(@"Unkown reason.", @"tracker stopped for unknown reason log message description");
		
			[self _writeTrackerStoppedToLogWithTitle:title description:desc];
			[_trackerController stopShowingTrackerProgress];
		}
	}
	else
		NSLog(@"caught fault getting tracker state, \n%@", [response fault]);	//+++ should show user.
		
	return self;
}


@end

@implementation TrackerDocument (Private)
- _lastDocumentPath;
{
	id		result = _lastDocumentPath;
	
	if (_lastDocumentPath == nil)
		result = [self fileName];
	
	return result;
}

- _setLastDocumentPath:path;
{
	[_lastDocumentPath release];
	_lastDocumentPath = [path copy];
	
	return self;
}
@end
