#!/usr/bin/env python
#
# Copyright (c) 2005, Sarwat Khan. All rights reserved.
# 
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
#         
#     1. Redistributions of source code must retain the above copyright
#        notice, this list of conditions and the following disclaimer.
#         
#     2. Redistributions in binary form must reproduce the above
#        copyright notice, this list of conditions and the following disclaimer
#        in the documentation and/or other materials provided with the
#        distribution.
#         
#     3. Neither Sarwat Khan's name nor the names of any contributors to
#        the BitTorrentAgent Software may be used to endorse or promote
#        products derived from this software without specific prior written
#        permission.
# 
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT
# OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#

## An abstraction for the BitTorrent python modules. Receives commands line-by-line, and produces XML as output.
## 
## The response dictionary contains the keys below. Note that in any objects, the 'fault' key is present only 
## if there was a failure. Any failure indicates a bug on your part (you sent a command with bad parameters).
## 
## In normal use you shouldn't get any faults.
## 	
## 	plist:
## 		dict:
## 			key: 	debug			#just useful info, rarely used. existence doesn't mean errors.
## 					array			#<- strings
## 					
## 			key: 	request			#the request string that was sent
## 					string
## 					
## 			key: 	data			#generic data result of any command
## 					object			#based on the command
## 			
## 			key:	fault			#request could not be performed for reasons in array
## 					array			#<- of any
## 										
## Stuff has changed completely with BitTorrent 4.0 (+ my own renovations in here). The best thing to
## do is to run TomatoTorrentPosixAgent.py on the command line and send it messages, trying to start a 
## download. Here's how. Assuming this directory layout,
## 
## 	bitconjurer-4/BitTorrent/
## 	Tomato/TomatoTorrentAgent/Python/
## 	
##  % export PYTHONPATH="bitconjurer-4:Tomato/TomatoTorrentAgent/Python"
##  % export TomatoCommandLineDebug="CUPCAKE"
##  % python Tomato/TomatoTorrentAgent/Python/TomatoTorrentPosixAgent.py
##  set metainfo sourceFile
##  /path/to/it.torrent
##  metainfo initialize
##  set download destPath
##  /path/to/something/with/final/path/name
##  set download configuration data_dir
##  /path/to/folder/to/hold/routing/table/stuff
##  download begin
##  
## Then, type 'download state' every now and then. To stop the download, type 'download end'. To shutdown
## the agent, type 'shutdown'. End of file doesn't quite cut it, but it will eventually get the message
## and shutdown itself.
##
## You need to set at least one valid configuration parameter, in this case I used max_upload_rate. Setting
## the second environment var makes it print the traceback rather than returning it as XML.
##
## Python takes a while to boot up and load all the modules. Type the 'vers' command first so you know it's ready.

from threading import *
import copy, sys, os, time, xml.sax.saxutils, traceback, base64, codecs
import pickle
import inspect

import BitTorrent
import BitTorrent.download, BitTorrent.track, BitTorrent.RawServer, BitTorrent.HTTPHandler, BitTorrent.parseargs, BitTorrent.bencode, BitTorrent.ConvertedMetainfo, BitTorrent.makemetafile, BitTorrent.defaultargs, BitTorrent.parseargs

from khashmir.utkhashmir import UTKhashmir
from khashmir import const

__all__ = ("Agent", "Metainfo", "Download", "setLoggingFunction", "copyright")

copyright = "Copyright (c) 2005, Sarwat Khan."

BitTorrent.makemetafile.ignore.append("Icon\r")
BitTorrent.makemetafile.ignore.append(".svn")

loggingFunction = None

def log_message(x):
	if loggingFunction: loggingFunction(x)

def setLoggingFunction(foo):
	global loggingFunction
	loggingFunction = foo

g_encoding = sys.getdefaultencoding()

##
##	PlistXMLWithObject
##
##		Output is parsed by Darwin's CoreFoundation. CF is open source
##		and compiles on most platforms (has native support for Windows). 
##		CF rocks.
##
##		It would be fairly easy to use something else to parse this XML,
##		just keep in mind that CF containers, like Python and Java containers,
##		can hold objects of any type (using the STL may not be convenient).
##
##		FreeBSD CF port: http://sarwat.net/opensource
##			An older version was integrated into Darwin; 
##			I'm unsure how up-to-date the current Darwin FreeBSD code is
##
##		There's also a pure python module to read and write this stuff 
##		at http://sarwat.net/opensource for Python 2.3 (not 2.2). The 'write'
##		code is better than this junk.

def PlistXMLWithObject(obj):
	
	def object_to_plist_xml(pyObject):
		result = ""
	
		if isinstance(pyObject, dict):
			result = dict_to_plist_xml(pyObject)
			
		elif isinstance(pyObject, list) or isinstance(pyObject, tuple):
			result = seq_to_plist_xml(pyObject)
		
		elif isinstance(pyObject, bool):
			if pyObject: result = "<true/>"
			else: result = "<false/>"

		elif isinstance(pyObject, long) or isinstance(pyObject, int):
			result = "<integer>%d</integer>" % pyObject
		
		elif isinstance(pyObject, float):
			result = "<real>%s</real>" % pyObject
			
		elif type(pyObject) == None.__class__:
			# the closest thing is NSNull, but that's not a plist type.
			# this shows up as <> as 'description', which is good.
			result = "<data></data>"	#not <data/>!
		
		elif isinstance(pyObject, unicode):
			result = u"<string>%s</string>" % xml.sax.saxutils.escape(pyObject)
			
		elif isinstance(pyObject, str):
			try:
				result = "<string>%s</string>" % xml.sax.saxutils.escape(unicode(pyObject, g_encoding))
			except:
				result = "<data>\n%s</data>" % base64.encodestring(pyObject)
		else:
			result = "<string>%s</string>" % xml.sax.saxutils.escape(str(pyObject).encode('ascii', 'replace'))
	
		return result
	
	def seq_to_plist_xml(pySeq):
		result = "<array></array>"
		
		count = len(pySeq)
		if count:
			newline = ""
			if count > 1: newline = "\n"
			
			result = "<array>" + newline
			
			for value in pySeq:
				result = result + object_to_plist_xml(value) + newline
			
			result = result + "</array>" + newline
		
		return result
	
	def dict_to_plist_xml(pyDict):
		result = "<dict></dict>"
	
		count = len(pyDict)
		
		if count:
			newline = ""
			if count > 1: newline = "\n"
			
			result = "<dict>" + newline
	
			for key in pyDict.keys():
				value = pyDict[key]
				
				plistKey = "<key>%s</key>" % key + newline
				plistValue = object_to_plist_xml(value) + newline
				
			
				result = result + plistKey + plistValue
			
			result = result + "</dict>" + newline
		
		return result
	
	
	result = u"""<?xml version="1.0" encoding="%s" ?>
<!DOCTYPE plist PUBLIC "-//Apple Computer//DTD PLIST 1.0//EN" "http://www.apple.com/DTDs/PropertyList-1.0.dtd">
<plist version="1.0">\n""" % g_encoding.upper()
	
	result = result + object_to_plist_xml(obj) + "\n"
	result = result + "</plist>\n"

	return result

##	Agent
##
##		command based front end to BitTorrent
##
##		supports one download instance and one tracker instance at a time. 
##		More downloads could easily be added, but it's just as easy to startup
##		another process (except for the python overhead). More trackers cannot
##		be added due to track.track() consuming sys.stdout.

class Agent (object):
	kLimbo = None
	kStartup = 1
	kRunning = 2
	kShutdown = 3
	
	def __init__(self, requests, responses):
		"Given input file requests and output file responses."
		
		utf8_e, utf8_d, utf8_sr, utf8_sw = codecs.lookup(g_encoding)
		requests = utf8_sr(requests, 'replace')
		# I'll do this myself, since I need to calculate an accurate length.
		#responses = utf8_sw(responses, 'replace')
		self._utf8_encode = utf8_e
		
		self._state = self.kLimbo
		self._requests = requests
		self._responses = responses	# there is also self._response = {}
		
		self._commands = { 	'set' 			: self.handleSet,			#reads a line for more info
							'get' 			: self.handleGet,			#useless, but whatever. 'data' in response
							'vars'			: self.handleVars,
							'clear'			: self.handleClear,
							'vers'			: self.handleVersion,		# BitTorrent.version
							'commands'		: self.handleCommands,	#query agent version through commands
							
							'output'		: self.handleOutput,		# takes token CFPropertyList or pickle
																		# defaults to CFPropertyList
																		
							'download'		: self.handleDownload,	#more tokens required from downloadCommands
							'tracker'		: self.handleTracker,
							'generate'		: self.handleGenerate,
							'metainfo'		: self.handleMetainfo,
							
							'shutdown'		: self.handleShutdown }

		self._metainfoCommands = {
							# vars[metainfo][sourceFile]
							# returns the converted metainfo dictionary. This also gets saved in vars[metainfo][convertedInfo].
							'initialize'	: self.handleMetainfoInitialize }
		
		self._downloadCommands = {
							'defaults'		: self.handleDownloadDefaults,
							'begin' 		: self.handleDownloadStart,	#includes state in response
							'state'			: self.handleDownloadState,	#state information regarding download
							'end'			: self.handleDownloadStop,	#nothing too useful on return
							'get'			: self.handleDownloadGetParameter,
							'set'			: self.handleDownloadSetParameter}

		self._trackerCommands = {
							'defaults'		: self.handleTrackerDefaults,
							'start' 		: self.handleTrackerStart,
							'stop'			: self.handleTrackerStop,
							'state'			: self.handleTrackerState}

		self._generateCommands = {
							#'defaults'		: self.handleGenerateDefaults,
							'start' 		: self.handleGenerateStart,
							'stop'			: self.handleGenerateStop,
							'state'			: self.handleGenerateState}
		
		self._outputCommands = {
							'CFPropertyList'	: self.handleOutputCF,
							'pickle'			: self.handleOutputPickle }	
							#can't 'get' output. Set what you want.
							#default is CFPropertyList

		self._commandTables = { 
							'request' : self._commands,
							'metainfo' : self._metainfoCommands,
							'download' : self._downloadCommands,
							'tracker' : self._trackerCommands,
							'generate' : self._generateCommands,
							'output' : self._outputCommands }

		# set sets vars into here. These are used for 'params' to BitTorrent.download
		
		# one of the vars that must be set is metainfo. 
		
		self._variables = {}
		
		self._response = None
		self._metainfo = None
		self._download = None
		self._tracker = None
		self._generate = None
		self.__output = PlistXMLWithObject
		
		self._num_noops = 0
		
		versionArray = BitTorrent.version.split('.')
		if len(versionArray) < 3: versionArray.append('0')
		
		self._btVers = dict(zip(['major', 'minor', 'fix'], [int(x) for x in versionArray]))
			
	#	run
	#	
	#		return value is None or an exception str.
	#
	#		Everything else in the class is private.
	
	def run(self):
		result = None
		
		self._state = self.kStartup
		self.startup()
		
		self._state = self.kRunning
		while self._state == self.kRunning:
			result = self.processNextRequest()

		self._state = self.kShutdown 		#pretty odd if this is necessary, but whatever.
		self.shutdown()

		return result
	
	def processNextRequest(self):
		"reads a line from requests, and sends it off to the right handler. " + \
		"Defines the response, which is turned into XML and then sent through the responses file"
		
		aborted = None
		self._response = {}
		self._response['fault'] = []
		self._response['debug'] = []
		
		try:
			request = self._requests.readline()
			self._response['request'] = request
			# log_message("TomatoTorrent.agent.Agent request: " + request)
			
			tokens = request.split()
			
			if self.__processTokensWithCommands(tokens, self._commands) == None:
				self._num_noops += 1
				self._response['debug'].append("empty request. num noops: %d" % self._num_noops)
				
				if self._num_noops > 3:
					self._state = self.kShutdown
					self._response['debug'].append("shutting down - too many empty commands")
					log_message("shutting down due to empty commands")
				
			else:
				self._num_noops = 0
				# signals and other interruptions can cause readline to abort.
			
			# and ship that response out.
			if len(self._response['fault']) == 0: del self._response['fault']
			self.sendResponse()
			
			
		except:
			# I really didn't think this through. Should put a dictionary to responses,
			# but responses' death could be the error itself. Whatever, just quit.
			try:
				# this is hopefully a broken pipe, in which case the except: below will trigger during sendResponse.
				if os.environ.get("TomatoCommandLineDebug") == None:
					aborted = str((sys.exc_info()[0], sys.exc_info()[1], traceback.format_tb(sys.exc_info()[2])))
					self._response['fault'] = aborted
					self.sendResponse()
				else:
					traceback.print_exc()
			except:
				if aborted == None:
					aborted = "Exception when trying to format exception."

				log_message("shutting down due to exception: %s" % aborted)
				self._state = self.kShutdown
		return aborted
		
	def startup(self):
		pass
		
			
	def shutdown(self):
		if self._download: 
			self._download.haltDownload()
			self._download.closeListeningSocket()

		if self._tracker:	self._tracker.stop()
		if self._generate:	self._generate.stop()
	
	def sendResponse(self):
		"Transforms the self._response dictionary into a CoreFoundation xml " + \
		"plist dictionary, and then appends it to the output"
		
		output = self.__output(self._response) + "\n"
		output, length_encoded = self._utf8_encode(output, 'replace')
		length = "%010d " % (len(output))				#exactly 11 bytes. I think \n is 2 on DOS.
		
		self._responses.writelines((str(length), str(output)))		#note that writelines doesn't add line separators.
		self._responses.flush()

		
	#	command handlers
	#
	#
	#
	def handleSet(self, tokens):
		"""Tokens are the path to the variable to set. Reads the next line of input to set the variable.
		
		The 'set download foo' would set self._variables['download']['foo']. Setting a variable and supplying
		no data unsets that variable (del's the key from the dictionary)."""
		
		if len(tokens): 

			variables, key = self.__dictionaryForKeyPath(tokens)
					
			value = self._requests.readline()
			if len(value) > 1:
				variables[key] = value[:-1]						#trailing newline
				self._response['data'] = variables[key]
				
			else:
				if variables.has_key(key):
					del(variables[key])
					self._response['debug'].append("removed " + str(key))

				#else:	
				#	self._response['fault'].append("no variable " + str(key) + " in path " + str(tokens))
				#
				#	not being able to unset an non-existant variable is no longer considered a failure.
				#	no key == no value == empty data
		else:
			self._response['fault'].append("no keys to set given")
	
	
	def handleGet(self, tokens):
		"Outputs the value of the given token in the 'data' field. Empty string if no var set."
		
		if len(tokens):
			data = ""
			
			variables, key = self.__dictionaryForKeyPath(tokens)
			if variables.has_key(key): 
				data = variables[key]
			
			self._response['data'] = data
			
		else:
			self._response['fault'].append("No variable to get given")
	

	def handleVars(self, tokens):
		self._response['data'] = self._variables
	

	def handleClear(self, tokens):
		variables, key = self.__dictionaryForKeyPath(tokens)
		variables.clear()
	
	def handleVersion(self, tokens):
		self._response['data'] = self._btVers
	
	def handleCommands(self, tokens):
	
		keys = self._commandTables.keys()
		result = dict(map(None, keys, map(lambda command : self._commandTables[command].keys(), keys)))
		self._response['data'] = result
	
	def handleOutput(self, tokens):
		
		if self.__processTokensWithCommands(tokens, self._outputCommands) == None:
			self.handleOutputCF(tokens)
	
	
	def handleOutputCF(self, tokens):
		self.__output = PlistXMLWithObject
	
	
	def handleOutputPickle(self, tokens):
		self.__output = pickle.dumps
	
	##
	##	Metainfo
	##
	##
	
	def handleMetainfo(self, tokens):
		if self.__processTokensWithCommands(tokens, self._metainfoCommands) == None:
			self._response['fault'].append("no download commands given")
	
	def handleMetainfoInitialize(self, tokens):
		"""If the meta info was usuable, self._metainfo is not None. 
		In any case, variables['metainfo']['convertedInfo'] will exist."""
		
		self._response['data'] = {}
		
		if self._variables.has_key('metainfo') \
			and isinstance(self._variables['metainfo'], dict) \
			and self._variables['metainfo'].has_key('sourceFile'):
			
			metainfoPath = self._variables['metainfo']['sourceFile']
			
			self._metainfo = Metainfo()
			self._metainfo.initializeFromPath(metainfoPath)
			
			stuff = self._metainfo.dictionaryRepresentation()
			stuff['source'] = metainfoPath
			self._response['data'] = stuff
			self._variables['metainfo']['convertedInfo'] = stuff
			
			if stuff['failure'] not in ('', None): self._metainfo = None
			else: del stuff['failure']
		else:
			self._response['fault'].append("Nothing to initialize a Metainfo from")

	##
	##	Download
	##
	##
	#
	#	Requirements:
	#		- self must have _metainfo created. See above.
	#		- self._variables['download']['destPath'] must be set.
	#		- self._variables['download']['configuration'] must be set. these are Download.defaults.
	#
	def handleDownload(self, tokens):

		if self.__processTokensWithCommands(tokens, self._downloadCommands) == None:
			self._response['fault'].append("no download commands given")
		

	def handleDownloadDefaults(self, tokens):
		def find_default(defaults, key):
			return [x for x in defaults if x[0] == key]
	
		defaults = BitTorrent.defaultargs.common_options# + BitTorrent.defaultargs.basic_options
		defaults = defaults + find_default(BitTorrent.defaultargs.rare_options, 'upload_unit_size')
		
		self._response['data'] = { 'defaults' : defaults,
								  'exceptions' : ('display_interval', 'data_dir', 'start_trackerless_client'),
								  'boolKeys' : ('check_hashes',) }
		
	def handleDownloadStart(self, tokens):
		"Spawns the thread to do the download. There is no useful response; you have to poll the state."
		
		if self._download == None \
			and self._variables.has_key('download') \
			and self._variables['download'].has_key('configuration') \
			and self._variables['download'].has_key('destPath') \
			and self._metainfo is not None:

			self._download = Download(self._variables['download']['configuration'], 
										self._metainfo.convertedMetainfo(), 
										self._variables['download']['destPath'])

			self._download.startDownload()
			self._response['data'] = {}	# don't really know what to return; they're 
										# going to have to poll to find out a state.
				
		else:
			self._response['fault'].append("can't start download. Either already downloading or no vars set.")
		
	def handleDownloadState(self, tokens):
		"Returns the current statistics. Especially useful if there's a download in progress."
		if self._download:
			state = self._download.state()
			self._response['data'] = state
			if state.has_key('fault'):	# this no longer happens, unless if I bring it back...
				self._response['fault'].append(state['fault'])
			
		else:
			self._response['fault'].append("no download started")
	
	def handleDownloadStop(self, tokens):
		"Stops the current download thread"
		
		if self._download:
			self._download.haltDownload()
			self._response['data'] = self._download.state()	# useless.... make them happy.
			self._download.closeListeningSocket()
			self._download = None
	
	def handleDownloadGetParameter(self, tokens):
		if self._download and len(tokens) == 1:
			parameter = tokens[0]
			self._response['data'] = {parameter:self._download.getParameter(parameter)}
		else:
			self._response['fault'].append('download get parameter called with bad parameters.') #whatever
			
	def handleDownloadSetParameter(self, tokens):
		if self._download and len(tokens) == 2:
			parameter, value = tokens
			self._download.setParameter(parameter, value)
			self._response['data'] = {parameter:self._download.getParameter(parameter)}
		else:
			self._response['fault'].append('download set parameter called with bad parameters.') #whatever
			
	##
	##	Tracker
	##	
	##
	def handleTracker(self, tokens):
		if self.__processTokensWithCommands(tokens, self._trackerCommands) == None:
			self._response['fault'].append("no tracker commands given")
		
	
	def handleTrackerDefaults(self, tokens):
		self._response['data'] = { 	'defaults' 	: BitTorrent.track.defaults,
									'exceptions': ('dfile', 'allowed_dir', 'logfile', 'port', 
														'min_time_between_log_flushes', 'hupmonitor'),
									'boolKeys'	: ('allow_get', 'parse_allowed_interval', 'show_names',
													'keep_dead', 'only_local_override_ip') }
									 					#port because I want to default it to 8080.
	
		
	def handleTrackerStart(self, tokens):
		if self._tracker == None and self._variables.has_key('tracker'):
			self._tracker = Tracker(self._variables['tracker'])

			data = self._tracker.start()

			if not data.has_key('fault'):
				self._response['data'] = data
				
			else:
				self._response['fault'].append("can't start tracker.")
				
				if data.has_key('fault'):
					self._response['fault'].append(data['fault'])

				self._tracker = None

		else:
			self._response['fault'].append("can't tracker download. Either already tracking or no vars set.")


	def handleTrackerStop(self, tokens):
		if self._tracker:
			data = self._tracker.stop()
			self._response['data'] = data
			self._tracker = None
			
		else:
			self._response['fault'].append("no tracker started.");
			

	def handleTrackerState(self, tokens):
		if self._tracker:
			data = self._tracker.state()
			self._response['data'] = data
			
			if data.has_key('fault'):
				self._response['fault'].append(data['fault'])
		else:
			self._response['fault'].append("no tracker started.")
			

	##
	##	Generate
	##
	##
	def handleGenerate(self, tokens):
		if self.__processTokensWithCommands(tokens, self._generateCommands) == None:
				self._response['fault'].append("no generate commands given")
		

	def handleGenerateDefaults(self, tokens):
		self._response['data'] = {	'defaults' 	: BitTorrent_generate.defaults,
									'exceptions'	: (),
									'boolKeys'		: () }
	
		
	def handleGenerateStart(self, tokens):
		if self._generate == None and self._variables.has_key('generate'):
			self._generate = Generate(self._variables['generate'])

			data = self._generate.start()

			if not data.has_key('fault'):
				self._response['data'] = data
				
			else:
				self._response['fault'].append("can't start generate.")
				
				if data.has_key('fault'):
					self._response['fault'].append(data['fault'])
				
				self._generate = None
		else:
			self._response['fault'].append("can't start generate. Either already generating or no vars set.")


	def handleGenerateStop(self, tokens):
		if self._generate:
			data = self._generate.stop()
			self._response['data'] = data
			self._generate = None
			
		else:
			self._response['fault'].append("generate not started.");
			

	def handleGenerateState(self, tokens):
		if self._generate:
			data = self._generate.state()
			self._response['data'] = data
			
			if data.has_key('fault'):
				self._response['fault'].append(data['fault'])
		else:
			self._response['fault'].append("no generate started.")
	

	def handleShutdown(self, tokens):
		"Sets the internal state to shutdown, which will cause shutdown() to be called and kill the download"
		
		self._state = self.kShutdown
	

	#	evaluates tokens against commands
	#
	def __processTokensWithCommands(self, tokens, commands):
		"Returns None if no commands given in tokens, but not if command was not recognized."

		result = None
		
		if len(tokens) > 0:
			result = 1
			if commands.has_key(tokens[0]):
				commands[tokens[0]](tokens[1:])
			else:
				self._response['fault'].append("command not recongized")

		return result


	#	for getting sub-dictionaries of self._variables
	#
	def __dictionaryForKeyPath(self, tokens):
		"Returns the dictionary in self._variables and the key as a tuple (dict, key)"

		result = self._variables, None
		
		if len(tokens):
			# initial dictionary to set
			variables = self._variables
	
			# get the dictionary for the path
			key_path = tokens[:-1]
			
			for key in key_path:
				if variables.has_key(key) != 1: variables[key] = {}
				variables = variables[key]
	
			#last item of tokens is the key to set in variables
			key = tokens[-1:][0]
			
			result = variables, key

		return result

##	agent.Metainfo
##
##		Used by TomatoTorrent.agent as an abstraction of convertedMetainfo. It's just a big dictionary.
class Metainfo (object):
	
	# these are results of self.status()
	def __init__(self):
		self.__convertedMetainfo = None
		self.__stuff = {'warnings':[], 'metainfo':None, 'failure':None, 'source':None}
		# metainfo in that is stuff translated from ConvertedMetaInfo.
	
	## initializers
	def initializeFromPath(self, path):
		"Initializes from a .torrent file."
		
		try:
			torrentFile = file(path)
			metainfoData = torrentFile.read()
			torrentFile.close()
			
			metainfo = BitTorrent.bencode.bdecode(metainfoData)
			convertedMetainfo = BitTorrent.ConvertedMetainfo.ConvertedMetainfo(metainfo)
			self.__metainfo = metainfo
			self.__convertedMetainfo = convertedMetainfo
			
			if convertedMetainfo.bad_torrent_wrongfield: self.__stuff['warnings'].append('encodedWrong')
			if convertedMetainfo.bad_torrent_unsolvable: self.__stuff['warnings'].append('hasNonUTF8Chars')
			if convertedMetainfo.bad_torrent_noncharacter: self.__stuff['warnings'].append('hasNonPrintingChars')
			if convertedMetainfo.bad_conversion: self.__stuff['warnings'].append('hasNamesBadForFileSystem')	# won't happen on a mac, it's UTF8.
			
			# taking stuff from ConvertedMetaInfo.
			computedKeys = 'origFiles', 'filesFs', 'name', 'nameFs', 'pieceLength', 'hashes', 'isArchive', 'sizes', 'totalBytes', 'comment', 'isTrackerless', 'announce', 'nodes'
			computedValues = convertedMetainfo.orig_files, convertedMetainfo.files_fs, convertedMetainfo.name, convertedMetainfo.name_fs, \
				convertedMetainfo.piece_length, convertedMetainfo.hashes, convertedMetainfo.is_batch, \
				convertedMetainfo.sizes, convertedMetainfo.total_bytes, metainfo.get('comment', ''), convertedMetainfo.is_trackerless, metainfo.get('announce', 'None'), metainfo.get('nodes', [])
			
			# this may not be a good idea for usability, but it's just for the GUI.
			# it doesn't have a need to be picky yet.
			foo = dict(self.__metainfo)
			foo.update(dict(zip(computedKeys, computedValues)))
			
			# zip up the sizes and files.
			if foo['filesFs'] is not None:
				thing = zip(foo['filesFs'], foo['sizes'])
				thing = [{'path':path, 'size':size} for path, size in thing]
				foo['archiveContents'] = thing
			
			self.__stuff['metainfo'] = foo
			self.__stuff['source'] = path
			
		except Exception, e:
			self.__stuff['failure'] = str(e)
		
		return self
		
	def dictionaryRepresentation(self):
		return self.__stuff
		
	def convertedMetainfo(self):
		return self.__convertedMetainfo
	
	def metainfoDictionary(self):
		# I don't think I care about this.
		return self.__metainfo

##	agent.Download
##
##		Used by TomatoTorrent.agent.Agent to handle a download. 
##		BTAgent only supports one download at a time right now.

class Download (object):

	defaults = {
		'display_interval': 0.05,
		'minport':6881,
		'maxport':6999,
		'bind':'',
		'rerequest_interval':5*60,
		'min_peers':20,
		'max_initiate':40,
		'max_upload_rate':20,
		'upload_unit_size': 1380,
		'max_uploads':-1,	# does something based on max_upload_rate
		'min_uploads':2,
		'max_files_open':50,
		'check_hashes':1,
		'forwarded_port':0,
		'ip':'',
		'max_allow_in':80,
		'data_dir':None}

	def __init__(self, params, convertedMetaInfo, filename):
		#
		#	This builds up everything and waits for 'startDownload' to fire.
		#	startDownload runs RawServer.listen_forever() in its own thread.
		#	This code configures RawServer so it's ready to run.
		#
		self.__errors = []
		self.__thread = None
		self.__stopSignal = Event()
		self.__metainfo = convertedMetaInfo
		self.__state = {}
		self.__stateMutex = Lock()
		
		# build up a config dictionary from BitTorrent.defaultargs, and then apply params on top of it.
		config = dict(self.defaults)
		config.update(dict(zip([x[0] for x in BitTorrent.defaultargs.rare_options], [x[1] for x in BitTorrent.defaultargs.rare_options])))
		config.update(params)
		
		if config.get('start_trackerless_client') == None:
			# we turn on the trackerless client if we have a trackerless torrent.
			config['start_trackerless_client'] = convertedMetaInfo.is_trackerless
		
		defaults = dict(Download.defaults)
		foo = list(BitTorrent.defaultargs.common_options)
		foo.extend(BitTorrent.defaultargs.rare_options)
		foo = [(x,y) for x,y,z in foo]
		defaults.update(dict(foo))
		BitTorrent.parseargs.parse_options(defaults, config)
		self.__params = config;
		self.__params['bad_libc_workaround'] = True
		
		runWithTrackerless = (self.__params.get('start_trackerless_client', None) != None)
		
		self.__displayInterval = self.__params.get('display_interval', self.defaults['display_interval'])
		
		# run loop
		self.__rawserver = BitTorrent.RawServer.RawServer(self.__stopSignal, 
									self.__params, 
									errorfunc=self._torrentError,
									tos=self.__params['peer_socket_tos'])
									
		# rate limiter and file pool
		self.__filepool = BitTorrent.Storage.FilePool(self.__params['max_files_open'])
		self.__ratelimiter = BitTorrent.RateLimiter.MultiRateLimiter(self.__rawserver.add_task)
		self.__ratelimiter.set_parameters(self.__params['max_upload_rate'], self.__params['upload_unit_size'])
		self.__dht = None

		# listen & dht. Find a suitable port.
		self.__listen = BitTorrent.Encoder.SingleportListener(self.__rawserver)
		if self.__params['minport'] <= 0: self.__params['minport'] = 1
		for port in xrange(self.__params['minport'], self.__params['maxport'] + 1):
			try:
				self.__listen.open_port(port, self.__params)
				if runWithTrackerless:
					self.__dht = UTKhashmir(self.__params['bind'], self.__listen.get_port(), 
											self.__params['data_dir'], self.__rawserver,
											int(self.__params['max_upload_rate'] * 1024 * 0.01),
											rlcount=self.__ratelimiter.increase_offset,	# I should look into why this is useful.
											config=self.__params)
				break
			except: pass
		else: 
			self.__errors.append("Could not open a listening port! Check your port range settings.")

		
		class Feedback (BitTorrent.download.Feedback):
			__slots__ = '_host'
			def __init__(self, host):					self._host = host
			def finished(self, torrent):				self._host._torrentFinished()
			def failed(self, torrent, garbageParam):	self._host._torrentFailed()
			def error(self, torrent, level, text):		self._host._torrentError(level, text)
			def started(self, torrent):					self._host._torrentStarted()
	
		# tie things together and wait for startDownload to be called.
		self.__torrent = BitTorrent.download._SingleTorrent(self.__rawserver, self.__listen, self.__ratelimiter,
															self.__filepool, self.__params, self.__dht)
		self.__torrent.rlgroup = BitTorrent.RateLimiter.RateLimitedGroup(self.__params['max_upload_rate'], self.__torrent.got_exception)
		self.__rawserver.add_context(self.__torrent)
		def start_torrent(): self.__torrent.start_download(self.__metainfo, Feedback(self), filename)
		self.__rawserver.add_task(start_torrent, 0, context=self.__torrent)

		self.__rawserver.external_add_task(self._fetchStats, self.__displayInterval)

	def startDownload(self):
		# go! I just don't want to spawn a thread in __init__.
		self.__thread = Thread(target=self.__rawserver.listen_forever)
		self.__thread.setDaemon(0)
		self.__thread.start()
		
	def haltDownload(self):
		def shutdown_torrent(): self.__torrent.shutdown
		self.__rawserver.external_add_task(shutdown_torrent, 0)
		self.__stopSignal.set()
	
	def closeListeningSocket(self):
		self.__listen.close_sockets()
	
	def state(self):
		"""Returns the state of the SingleTorrent. 
		
		This includes downtotal, uptotal, and status,
		which is a bunch of stuff from DownloaderFeedback."""
		result = {}
		
		self.__stateMutex.acquire() ##
		result.update(self.__state)
		self.__stateMutex.release() ##

		return result
	
	def setParameter(self, option, value):
		value = int(value)
		if option == 'max_files_open':
			self.__filepool.set_max_files_open(value)
		elif option == 'max_upload_rate':
			foo, bar = value, self.__params['upload_unit_size']
			self.__ratelimiter.set_parameters(foo, bar)
			self.__torrent.set_option(option, value)
		elif option == 'upload_unit_size':
			foo, bar = self.__params['max_upload_rate'], value
			self.__ratelimiter.set_parameters(foo, bar)
	
	def getParameter(self, option):
		result = None
		if option == 'max_files_open':
			result = self.__filepool.max_files_open
		elif option == 'max_upload_rate':
			result = self.__ratelimiter.upload_rate
			if result > 0: result = result / 1024
		elif option == 'upload_unit_size':
			result - self.__ratelimiter.unit_size
		return result
	
	def _torrentStarted(self):
		log_message("Torrent Started.")

	def _torrentFinished(self):
		log_message("Torrent Finished.")
		
	def _torrentFailed(self):
		# raw server got an exception. everything's going to go away now.
		log_message("Torrent Failed")
	
	def _torrentError(self, level, text):
		log_message("Torrent Error: %s, %s, %s" % (time.time(), level, text))
		fudge = (time.time(), text)
		self.__errors.append(fudge)	##atomic.

	def _fetchStats(self):
		# hooked up to the raw server. fires every self.__displayInterval
		status = self.__torrent.get_status()

		uptotal, downtotal = self.__torrent.get_total_transfer()
		errors = list(self.__errors)

		self.__stateMutex.acquire() ##
		self.__state = {'status':status, 'downtotal':downtotal, 'uptotal':uptotal, 'errors':errors}
		
		activity = status.get('activity', '')
		timeEst = status.get('timeEst', 0)
		if activity == 'downloading' and timeEst > 0 and timeEst < 5184000: #60 days
			hours, r = divmod(timeEst, 60 * 60)
			minutes, seconds = divmod(r, 60)
			if hours < 1000000:
				status['hmsTimeEstimate'] = (hours, minutes, seconds)

		self.__state['status']['TomatoDownloadSizeBytes'] = self.__metainfo.total_bytes;

		self.__stateMutex.release() ##
		
		# reschedule.
		self.__rawserver.external_add_task(self._fetchStats, self.__displayInterval)
	

class Tracker (object):
	#Should have made a base class. oh well.
	#Intended to be used with a log file. Else the output will go to stderr (stdout set in main routine).
	
	kLimbo = "limbo"					## BitTorrent.download.download() not running
	kRunning = "running"				## state set before calling download()
	kFailed = "failed"
	
	def __init__(self, params):
		
		self._args = self.__initArgsFromDictionary(params)
		self._trackerState = {}
		self._thread = None
		self.stateMutex = Lock()
		self._stopEvent = None
		
		self.__setStatus(self.kLimbo)
	
	
	def start(self):
		if len(self._args) > 0 and self._thread == None:
			try:
				self.config, self.files = BitTorrent.parseargs.parseargs(self._args, BitTorrent.track.defaults, 0, 0)

				self._stopEvent = Event()
				self.__setStatus(self.kRunning)

				self._thread = Thread(target=self._trackerThread)
				self._thread.start()
				
				self.logflush = Thread(target=self._flushLog)
				self.logflush.start()
				
			except:
				self.__setStatus(self.kFailed, (str(sys.exc_info()[0]), str(sys.exc_info()[1])))
		else:
			self.__setStatus(self.kFailed, "already running or no arguments given.")
			
		return self.state()
	
	
	def stop(self):
		if self._thread != None:
			self._stopEvent.set()
			self._thread = None
			
		return self.state()

	
	def state(self):
		result = {}
		
		self.stateMutex.acquire() ##
		
		try: result = copy.copy(self._trackerState)
		except: pass
			
		self.stateMutex.release() ##

		return result

	def _trackerThread(self):
		try:
			oldstdout = sys.stdout
			
			r = BitTorrent.RawServer.RawServer(self._stopEvent, self.config)
			t = BitTorrent.track.Tracker(self.config, r)
			s = r.create_serversocket(self.config['port'], self.config['bind'], True)
			r.start_listening(s, BitTorrent.HTTPHandler.HTTPHandler(t.get, self.config['min_time_between_log_flushes']))
			r.listen_forever()

			t.save_dfile()
			self.__setStatus(self.kLimbo)
			
			if sys.stdout != oldstdout:
				try:
					sys.stdout.close()
				except:
					pass
				
				sys.stdout = oldstdout
			
		except:
			self.__setStatus(self.kFailed, (str(sys.exc_info()[0]), str(sys.exc_info()[1])))
	
	def _flushLog(self):
		# i see now that there's a flush log interval param. oh well.
		stopEvent = self._stopEvent
		
		while not stopEvent.isSet():
			try:
				self._stopEvent.wait(1)
				sys.stdout.flush()
			except:
				sys.stderr.write(str((str(sys.exc_info()[0]), str(sys.exc_info()[1]))) + "\n")
				

	def __setStatus(self, status, fault = None):

		self.stateMutex.acquire()	##

		self._trackerState['status'] = status
		self._trackerState['statusSetAtTime'] = time.time()
		
		if fault:
			self._trackerState['fault'] = fault
		elif self._trackerState.has_key('fault'):
			del self._trackerState['fault']
			
		self.stateMutex.release()	##

		
	def __initArgsFromDictionary(self, params):
		self._args = []
		
		for key in params.keys():
			value = params[key]
			
			self._args.append("--" + str(key))
			self._args.append(str(value))
		return self._args	#yes, this is pointless. It just makes 'args' existence clear in the init method.

		

class Generate (object):
	#Will the wonders of copy and paste ever cease?
	
	kLimbo = "limbo"
	kFinished = "finished"			# after starting the state is not limbo, 
	kFailed = "failed"				# it's one of these
	kRunning = "running"
	
	def __init__(self, params):
		
		###	params:
		#
		#		_agent_files:			a dictionary of files to use. keys are ignored; value is file path
		#		_agent_filesAttribs:	(optional) a dictionary of attributes for each file, as defined below.
		#								you don't have to specify a dictionary for each file. keys are ignored,
		#								but have to equal the key for the file in _agent_files.
		#		_agent_destDir			default directory to write each torrent in.
		#		_agent_trackerURL		default tracker url
		#		_agent_trackerless		(optional) YES if generating trackerless.
		#		_agent_dht_nodes		(optional) comma-separated list of host:port nodes
		#		data_dir				(optional) where the routing table is.
		#		piece_size_pow2:		(optional) piece size for generate.make_meta_file. Default is 18.
		#
		#		_agent_filesAttribs values:
		#			trackerURL			(optional) tracker url
		#			comment				(optional) file comment
		#			destDir				(optional) destination folder for torrent
		#
		#		example:
		#			set generate _agent_destDir
		#			/blah
		#			set generate _agent_trackerURL
		#			http://default.tracker/announce
		#			set generate _agent_files 0
		#			/path/to/file.text
		#			set generate _agent_filesAttribs 0 trackerURL
		#			http://tracker.url.for/file.text.torrent
		
		#all member vars except generateState are used by the thread only.
		self.destDir = params['_agent_destDir']
		self._trackerURL = params['_agent_trackerURL']
		self.pieceSize = long(params.get('piece_size_pow2', 0))
		if self.pieceSize == 0: self.pieceSize = 18	
		
		#regarding the existing code in this object: it's dumb.
		self._trackerless = (params.get("_agent_trackerless", "NO") == "YES")
		self._dhtNodes = str(params.get("_agent_dht_nodes", ""))
		self._dataDir = params.get("data_dir", None)

		#self.files is key=path, value=attributes for path as defined in _agent_filesAttribs, or {} if not def'd.
		files = params['_agent_files']
		attribs = params.get('_agent_filesAttribs', {})
		self.files = dict(zip([str(x) for x in files.values()], #de-unicode files.values() now
								[attribs.get(key, {}) for key in files.keys()]))

		self.stateMutex = Lock()
		self._generateState = {'errors': []}

		self._thread = None
		self._stopEvent = None
		
		self.__setStatus(self.kLimbo)
	
	def start(self):
		if len(self.files) > 0 and self._thread == None:
			try:
				self._stopEvent = Event()
				self.__setStatus(self.kRunning)

				self._thread = Thread(target=self._generateThread)
				self._thread.start()
				
			except:
				self.__setStatus(self.kFailed, (str(sys.exc_info()[0]), str(sys.exc_info()[1])))
		else:
			self.__setStatus(self.kFailed, "already running or no arguments given.")
			
		return self.state()
	
	
	def stop(self):
		if self._thread != None:
			self._stopEvent.set()
			self._thread = None
			
		return self.state()

	
	def state(self):
		result = {'status' : "unknown"}
		
		self.stateMutex.acquire() ##
		try: result = copy.copy(self._generateState)
		except: pass
			# I'm the one who sets all the data in generateState,
			# so as of this writing deepcopy is not going to fail.
		self.stateMutex.release() ##

		return result

	def _generateThread(self):
		try:
			
			#isn't this a lovely line of code. It just kept growing. 
			#	join(directory, name + .torrent) for each file
			torrents = [os.path.join(self.files[file].get('destDir', self.destDir), os.path.split(file)[-1] + ".torrent") \
							for file in self.files.keys()]
			sizes = [BitTorrent.makemetafile.calcsize(file) for file in self.files]

			self.stateMutex.acquire() ##
			self._generateState['numFiles'] = len(torrents)
			self._generateState['totalBytes'] = reduce(lambda a, b: a+b, sizes)
			self._generateState['completedBytes'] = 0
			totalBytesProcessed = 0
			
			for path, size, torrent, index in map(None, self.files.keys(), sizes, torrents, range(len(torrents))):
				
				self._generateState['currentItemPath'] = path
				self._generateState['currentItemNum'] = index+1
				self.stateMutex.release() ##

				#abort condition
				if self._stopEvent.isSet(): break
				
				attribs = self.files[path]
				tracker = attribs.get('trackerURL', self._trackerURL)
				comment = attribs.get('comment', None)

				# un-unicode them.
				if comment: comment = str(comment)
				tracker = str(tracker)
				
				fileName = os.path.split(path)[-1]
				# we want ds-store files.
				if (fileName not in BitTorrent.makemetafile.ignore and fileName[0] != '.'):
					try:
						if not self._trackerless:
							BitTorrent.makemetafile.make_meta_file(
								path, tracker, target=torrent,
								flag = self._stopEvent, 
								progress = self._progress, 
								piece_len_exp = self.pieceSize,
								comment = comment,
								encoding = g_encoding)
						else:
							BitTorrent.makemetafile.make_meta_file_dht(
								path, self._dhtNodes, target=torrent,
								flag = self._stopEvent,
								progress = self._progress,
								piece_len_exp = self.pieceSize,
								comment = comment,
								encoding = g_encoding,
								data_dir = self._dataDir)
						
						self.stateMutex.acquire() ##
						totalBytesProcessed += size
						self._generateState['completedBytes'] = totalBytesProcessed
						
					except:
						self.stateMutex.acquire() ##
						self._generateState['errors'].append((torrent, (str(sys.exc_info()[0]), str(sys.exc_info()[1]))))
						self._generateState['completedBytes'] = totalBytesProcessed
						self._generateState['totalBytes'] -= size
			
				
			self.stateMutex.release() ##
			
			self.__setStatus(self.kFinished)

		except:
			self.__setStatus(self.kFailed, (str(sys.exc_info()[0]), str(sys.exc_info()[1])))
	
	def _progress(self, numBytes):

		self.stateMutex.acquire() ##
		self._generateState['completedBytes'] += numBytes
		self.stateMutex.release() ##
		

	def __setStatus(self, status, fault = None):

		self.stateMutex.acquire()	##

		self._generateState['status'] = status
		self._generateState['statusSetAtTime'] = time.time()
		
		if fault:
			self._generateState['fault'] = fault
		elif self._generateState.has_key('fault'):
			del self._generateState['fault']
			
		self.stateMutex.release()	##
		
