//==============================================================================
//		TomatoTorrentAgent.m					TomatoTorrent
//==============================================================================
//  Created by Sarwat Khan on Tue Jun 03 2003.
//  Copyright (c) 2003, Sarwat Khan. All rights reserved.
//  
//  Redistribution and use in source and binary forms, with or without
//  modification, are permitted provided that the following conditions are
//  met:
//          
//      1. Redistributions of source code must retain the above copyright
//         notice, this list of conditions and the following disclaimer.
//          
//      2. Redistributions in binary form must reproduce the above
//         copyright notice, this list of conditions and the following disclaimer
//         in the documentation and/or other materials provided with the
//         distribution.
//          
//      3. Neither Sarwat Khan’s name nor the names of any contributors to
//         the Macintosh BitTorrent Software may be used to endorse or promote
//         products derived from this software without specific prior written
//         permission.
//  
//  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
//  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
//  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
//  A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT
//  OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
//  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
//  LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
//  DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
//  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
//  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
//  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


#import "TomatoTorrentAgent.h"
#import "tomato_agent_logging.h"
#import "AgentLogging.h"

#import "SKRange.h"
#import "NSDictionary+DataTransform.h"

#import "tomato_agent.h"

@interface TomatoTorrentAgent (Private)
- __beginDownloadStrings:destPath;
- __beginTrackerStrings;
- __beginGenerateStrings;
- __queueTransaction:transaction forClient:sender;
- __startTaskThread;
- __stopTaskThread;

- (void) __taskThread:param;
- __taskLoop;
- __taskGatewayLoop;
- __processTransaction:transaction;

+ (void) __runTaskToFetchDefaults:mutex;
@end


//keys used internally when running the task. These are excluded from the defaults/options dictionary.
#define kBTDownload_responseKey							@"responsefile"
#define kBTDownload_urlKey								@"url"

#define kDefaultsAnswerKey_defaults						@"defaults"
#define kDefaultsAnswerKey_exceptions					@"exceptions"
#define kDefaultsAnswerKey_boolKeys						@"boolKeys"

#define kTransactionQueueEmpty							0
#define kQueueHasTransactions							1

#define kShutdownLock_running							0
#define kShutdownLock_shutdown							1


#define kDownloadStateTransactionName_					NSStringFromSelector(@selector(downloadState:))
#define kDownloadConfigTransactionName_					NSStringFromSelector(@selector(downloadConfiguration:))

//------------------------------------------------------------------------------
//		• globals
//------------------------------------------------------------------------------

static id g_globals_lock = nil;
static id g_download_defaults = nil;
static id g_tracker_defaults = nil;
static id g_torrent_data_dir = @"";

//------------------------------------------------------------------------------
//		• local functions
//------------------------------------------------------------------------------

static id TTAgentPythonLibPath(Class agent);

//------------------------------------------------------------------------------
//		• TomatoTorrentAgentTransaction
//------------------------------------------------------------------------------

@interface TomatoTorrentAgentTransaction: NSObject
{
	id		__target;		//receives the result of transaction. may be nil.
	SEL		__action;		//selector to perform. Gets performed in main run loop.
	id		__name;
	id		__data;			//array of strings to send the task as a commands/data.
	int		__count;			//effective number of transactions object represents
}
- initWithStrings:data name:name resultsFor:target selector:(SEL)sel;
- strings;
- postResponse:response forAgent:agent;
- (unsigned) effectiveCount;
- incrementCount;
@end

@interface TomatoTorrentAgentResponse (TomatoTorrentAgent)
- initWithResponseDictionary:data respondingTo:name;
- setConversationDuration:(double)duration;
@end


//------------------------------------------------------------------------------
//		• TomatoTorrentAgent
//------------------------------------------------------------------------------


@implementation TomatoTorrentAgent
+ (void) initialize
{
	bool foo = false; if (foo) return; foo = true;
	
	ttagent_initialize(
		[[[NSBundle bundleForClass:[self class]] pathForResource:@"pythonInfo" ofType:@"plist"] UTF8String]);
	g_globals_lock = [NSLock new];
	
	g_torrent_data_dir = [NSTemporaryDirectory() copy];
}

- init
{
	[super init];
	
	__displayInterval = 0.75;	//TomatoTorrentAgent.py will internally refresh
								//stats download at this interval
	__trackerPort = 8080;
	
	__agentLock = [[NSLock allocWithZone:[self zone]] init];	
	__shutdownLock = [[NSConditionLock allocWithZone:[self zone]] 
		initWithCondition:kShutdownLock_running];
	
	[self __startTaskThread];
	
	return self;
}

- (void) dealloc
{
	[__downloadConfiguration release];
	[__trackerOptions release];
	[__trackerAllowedDirPath release];
	[__trackerLogFilePath release];
	[__trackerDfile release];
	[__generateOptions release];
	
	[__shutdownLock release];
	[__agentLock release];
	
	[super dealloc];
}


- mutex
{
	return __agentLock;		//see header.
}

- shutdown
{
	[self __stopTaskThread];
	return self;
}

//------------------------------------------------------------------------------
//		• defaultDownloadConfigurationDictionaryList
//------------------------------------------------------------------------------
//	Returns the array of options from BitTorrent.download.defaults, minus the
//	keys excluded by the agent (not generically user-configurable).
//
//	This looks complicated because the main thread shouldn't get stuck if it's impossible
//	to get the defaults. It spawns a thread to get the defaults and gives up on it after
//	15 seconds (note that python does take a couple seconds to start up too).
//
+ defaultDownloadConfigurationDictionary
{
	id				result = nil;
	static BOOL 	failed = NO;
	
	[g_globals_lock lock];
	{
		if (g_download_defaults == nil && failed == NO)
		{
			//run the task to define it.
			id		mutex, params;
			
			mutex = [[NSConditionLock alloc] initWithCondition:0];
			params = [NSMutableDictionary dictionaryWithObjectsAndKeys:
				mutex, 					@"mutex",
				@"download defaults", 	@"command", 
				nil];

			[NSThread detachNewThreadSelector:@selector(__runTaskToFetchDefaults:) 
									 toTarget:self 
								   withObject:params];
			
			if ([mutex lockWhenCondition:1 beforeDate:[NSDate dateWithTimeIntervalSinceNow:15]])
			{
				result = [[params objectForKey:@"defaults"] retain];	//nil if failed.
				[mutex unlock];
				[mutex release];
			}
			else
			{
				failed = YES;
				result = [NSDictionary dictionaryWithObjectsAndKeys:
					[NSArray array],			kTomatoAgentModuleDefaultsKey_list,
					[NSDictionary dictionary],	kTomatoAgentModuleDefaultsKey_map,
					[NSArray array],			kTomatoAgentModuleDefaultsKey_boolKeys,
					nil];
				[result retain];
			}
			
			g_download_defaults = result;
		}
		
		result = g_download_defaults;
	}
	[g_globals_lock unlock];
	
	return result;
}

- defaultDownloadConfigurationDictionary
{
	return [[self class] defaultDownloadConfigurationDictionary];
}

//------------------------------------------------------------------------------
//		• defaultTrackerOptions
//------------------------------------------------------------------------------
//	look familiar?

+ defaultTrackerOptions
{
	id				result = nil;
	static BOOL 	failed = NO;
	
	[g_globals_lock lock];
	{
		if (g_tracker_defaults == nil && failed == NO)
		{
			//run the task to define it.
			id		mutex, params;
			
			mutex = [[NSConditionLock alloc] initWithCondition:0];
			params = [NSMutableDictionary dictionaryWithObjectsAndKeys:
				mutex, 					@"mutex",
				@"tracker defaults", 	@"command", 
				nil];
			
			[NSThread detachNewThreadSelector:@selector(__runTaskToFetchDefaults:) 
									 toTarget:self 
								   withObject:params];
			
			if ([mutex lockWhenCondition:1 beforeDate:[NSDate dateWithTimeIntervalSinceNow:15]])
			{
				result = [[params objectForKey:@"defaults"] retain];	//nil if failed.
				[mutex unlock];
				[mutex release];
			}
			else
			{
				failed = YES;
				result = [NSDictionary dictionaryWithObjectsAndKeys:
					kTomatoAgentModuleDefaultsKey_list,		[NSArray array],
					kTomatoAgentModuleDefaultsKey_map,		[NSDictionary dictionary],
					kTomatoAgentModuleDefaultsKey_boolKeys,	[NSArray array],

					nil];
				[result retain];
			}
			
			g_tracker_defaults = result;
		}
		
		result = g_tracker_defaults;
	}
	[g_globals_lock unlock];
	
	return result;
}

- defaultTrackerOptions
{
	return [[self class] defaultTrackerOptions];
}

//------------------------------------------------------------------------------
//		• __validateAgentOptions:againstDefaultsSelectedBy:
//------------------------------------------------------------------------------
//	Removes any keys that are no longer in default options, and adds any keys
//	that are in defaults options that are not in options. options is a mutable
//	dictionary.

+ __validateAgentOptions:options againstDefaultsSelectedBy:(SEL)defaultsSelector;
{
	id	defaults;
	id	key, ite;
	
	
	defaults = [[self performSelector:defaultsSelector] objectForKey:kTomatoAgentModuleDefaultsKey_map];

	//need to add any new keys in defaults that are not in options
	ite = [defaults keyEnumerator];
	while (key = [ite nextObject])
	{
		if ([options objectForKey:key] == nil)
		{
			[options setObject:[[defaults objectForKey:key] objectAtIndex:kTomatoAgentModuleDefaultsEntryIndex_value]
						forKey:key];
		}
	}
	
	//need to remove any keys in options that are not in defaults
	ite = [options keyEnumerator];
	while (key = [ite nextObject])
	{
		if ([defaults objectForKey:key] == nil)
			[options removeObjectForKey:key];
	}
	
	return self;
}

//------------------------------------------------------------------------------
//		• downloadOptions
//------------------------------------------------------------------------------
//	Returns mutable dictionary of defaults, where the keys are from the py
//	module BitTorrent.download.defaults. Be aware of thread-safety when modifying
//	the dictionary (ie there should only ever be one thread that modifies it, else
//	put a lock to access this thing. Also it uses NSUserDefaults which isn't thread-safe).

static id downloadOptions_map(id object, id key, void *context)
{
	return [object objectAtIndex:kTomatoAgentModuleDefaultsEntryIndex_value];
}

- downloadConfigurationDictionary;
{
	if (__downloadConfiguration == nil)
	{
		BOOL	shouldValidate = NO;

		//init from defaults
		__downloadConfiguration = [[NSUserDefaults standardUserDefaults] 
			dictionaryForKey:kUserDefault_TomatoAgent_downloadOptions];
		
		if (__downloadConfiguration == nil)
			//no defaults, use the defaultDownloadConfigurationDictionary, mapping it to contain only values for each key
			__downloadConfiguration = 
				[[[self defaultDownloadConfigurationDictionary] objectForKey:kTomatoAgentModuleDefaultsKey_map]
					dictionaryMappingObjectsUsingFunction:downloadOptions_map context:nil];
		else
			shouldValidate = YES;
			
		//make a mutable dictionary out of it.
		__downloadConfiguration = [[NSMutableDictionary allocWithZone:[self zone]] 
								initWithDictionary:__downloadConfiguration];

		if (shouldValidate)
			[TomatoTorrentAgent __validateAgentOptions: __downloadConfiguration 
							 againstDefaultsSelectedBy: @selector(defaultDownloadConfigurationDictionary)];
	}
	return __downloadConfiguration;
}

- setDownloadConfigurationDictionary:foo
{
	[__downloadConfiguration release];
	__downloadConfiguration = [[NSMutableDictionary dictionaryWithDictionary:foo] retain];
	
	return self;
}

+ (void) setTorrentCachesDirectory:(NSString*)dir;
{
	g_torrent_data_dir = [dir copy];
}

+ (NSString*) torrentCachesDirectory;
{
	return g_torrent_data_dir;
}

//------------------------------------------------------------------------------
//		• trackerOptions
//------------------------------------------------------------------------------
//	Returns mutable dictionary of defaults, where the keys are from the py
//	module BitTorrent.tracker.defaults. Be aware of thread-safety when modifying
//	the dictionary (ie there should only ever be one thread that modifies it, else
//	put a lock to access this thing. Unlike downloadOptions, this does not use NSUserDefaults.
//
//	The trackerOptions can be set using setTrackerOptions, which is why it doesn't
//	use NSUserDefaults.

static id trackerOptions_map(id object, id key, void *context)
{
	return [object objectAtIndex:kTomatoAgentModuleDefaultsEntryIndex_value];
}

- trackerOptions;
{
	if (__trackerOptions == nil)
	{
		[self setTrackerOptions:[
			[[self defaultTrackerOptions] objectForKey:kTomatoAgentModuleDefaultsKey_map]
				dictionaryMappingObjectsUsingFunction:trackerOptions_map context:nil]];
	}	

	return __trackerOptions;
}

- setTrackerOptions:dictionary
{
	[__trackerOptions release];
	__trackerOptions = [[NSMutableDictionary allocWithZone:[self zone]] initWithDictionary:dictionary];
	[TomatoTorrentAgent __validateAgentOptions:__trackerOptions againstDefaultsSelectedBy:@selector(defaultTrackerOptions)];
	return self;
}

- setStatsUpdateInterval:(double)value
{
	__displayInterval = value;
	return self;
}

- (double) statsUpdateInterval
{
	return __displayInterval;
}

- (BOOL) isRunning
{
	return [__shutdownLock condition] == kShutdownLock_running;
}

- setTrackerPort:(int)port;
{
	__trackerPort = port;
	return self;
}

- setTrackerLogFilePath:path;
{
	[__trackerLogFilePath release];
	__trackerLogFilePath = [path retain];
	return self;
}

- setTrackerStateFilePath:path;
{
	[__trackerDfile release];
	__trackerDfile = [path retain];
	return self;
}

- setTrackerAllowedDirPath:path;
{
	if (path == nil || [path isAbsolutePath])
	{
		[__trackerAllowedDirPath release];
		__trackerAllowedDirPath = [path retain];
	}
	
	return self;
}

- (int)trackerPort;
{
	return __trackerPort;
}

- trackerLogFilePath;
{
	return __trackerLogFilePath;
}

- trackerStateFilePath;
{
	return __trackerDfile;
}


- trackerAllowedDirPath
{
	return __trackerAllowedDirPath;
}

//------------------------------------------------------------------------------
//		• generateOptions
//------------------------------------------------------------------------------
//	Initially empty. You must use setGenerateAnnounceURL:, setGenerateDestDirPath:,
//	and setGenerateFiles:. This will set the corresponding keys in generateOptions.

- generateOptions;
{
	if (__generateOptions == nil)
		__generateOptions = [[NSMutableDictionary allocWithZone:[self zone]] init];
	
	return __generateOptions;
}

- setGenerateAnnounceURL:string;
{
	[[self generateOptions] setObject:string forKey:kTomatoAgentKey_generate_trackerURL];
	return self;
}

- setGenerateDestDirPath:string;
{
	[[self generateOptions] setObject:string forKey:kTomatoAgentKey_generate_destDirPath];
	return self;

}

- setGenerateTrackerless:(BOOL)foo;
{
	[[self generateOptions] setObject:(foo ? @"YES" : @"NO") forKey:kTomatoAgentKey_generate_isTrackerless];
	return self;
}

- setGenerateDHTNodesString:nodes;
{
	[[self generateOptions] setObject:[[nodes copy] autorelease] forKey:kTomatoAgentKey_generate_DHTNodes];
	return self;
}

- setGeneratePieceSize:(int)size;
{
	[[self generateOptions] setObject:[NSNumber numberWithInt:size] forKey:kTomatoAgentKey_generate_pieceSizePow2];
	return self;
}

- setGenerateFiles:paths;
{
	[[self generateOptions] setObject:paths
							   forKey:kTomatoAgentKey_generate_files];
	return self;
}

- setGenerateFiles:paths withAttributes:dictionaries;
{
	[[self generateOptions] setObject:paths
							   forKey:kTomatoAgentKey_generate_files];
			 
	[[self generateOptions] setObject:dictionaries
							   forKey:kTomatoAgentKey_generate_filesAttribs];
					 		 
	return self;
}

- generateAnnounceURL;
{
	return [[self generateOptions] objectForKey:kTomatoAgentKey_generate_trackerURL];
}

- generateDestDirPath;
{
	return [[self generateOptions] objectForKey:kTomatoAgentKey_generate_trackerURL];
}

- (int)generatePieceSize;
{
	return [[[self generateOptions] objectForKey:kTomatoAgentKey_generate_pieceSizePow2] intValue];
}

- generateFiles;
{
	return [[self generateOptions] objectForKey:kTomatoAgentKey_generate_files];
}

- (NSArray*) generateFilesAndAttributes;	//result contains two objects, files and attribute dictionaries
{
	id		files, attribs;
	
	files = [self generateFiles];
	attribs = [[self generateOptions] objectForKey:kTomatoAgentKey_generate_filesAttribs];
	
	if (files == nil) files = [NSNull null];
	if (attribs == nil) attribs = [NSNull null];
	
	return [NSArray arrayWithObjects:files, attribs, nil];
}


//------------------------------------------------------------------------------
//		 • messaging
//------------------------------------------------------------------------------
//	This is asyncronous. When you send a message, you will get a response
//	eventually. You will get the response directly if you respond to the message's
//	response selector. You can get a response indirectly by listening to notifications
//	coming from this object. The name of the notification is the name of the message
//	(the selector/method name that you invoke).

- metainfoPath:path delegate:sender;
{
	id		transaction;
	id		target = nil;
	SEL		sel = @selector(tomatoTorrentAgent:metainfo:);
	
	if ([sender respondsToSelector:sel])
		target = sender;
	
	transaction = [[TomatoTorrentAgentTransaction allocWithZone:[self zone]] 
		initWithStrings: [NSArray arrayWithObjects:
							[NSString stringWithFormat:@"set metainfo sourceFile\n%@\n", path],
							@"metainfo initialize",
							nil]
				   name: NSStringFromSelector(_cmd)
			 resultsFor: target 
			   selector: sel];
	
	return [self __queueTransaction:[transaction autorelease] forClient:sender];
}

- metainfo:sender;
{
	id		transaction;
	id		target = nil;
	SEL		sel = @selector(tomatoTorrentAgent:metainfo:);
	
	if ([sender respondsToSelector:sel])
		target = sender;
	
	transaction = [[TomatoTorrentAgentTransaction allocWithZone:[self zone]] 
		initWithStrings: [NSArray arrayWithObject:@"metainfo"]
				   name: NSStringFromSelector(_cmd)
			 resultsFor: target 
			   selector: sel];
	
	return [self __queueTransaction:[transaction autorelease] forClient:sender];
}

- startDownloadingToPath:path delegate:sender;
{
	id		transaction;
	id		target = nil;
	SEL		sel = @selector(tomatoTorrentAgent:downloadStartup:);
	
	if ([sender respondsToSelector:sel])
		target = sender;
	
	transaction = [[TomatoTorrentAgentTransaction allocWithZone:[self zone]] 
		initWithStrings:[self __beginDownloadStrings:path]
				   name:NSStringFromSelector(_cmd)
			 resultsFor:target 
			   selector:sel];
	
	return [self __queueTransaction:[transaction autorelease] forClient:sender];
}

- endDownload:sender
{
	id		transaction;
	id		target = nil;
	SEL		sel = @selector(tomatoTorrentAgent:downloadEnded:);
	
	if ([sender respondsToSelector:sel])
		target = sender;
	
	transaction = [[TomatoTorrentAgentTransaction allocWithZone:[self zone]] 
		initWithStrings:[NSArray arrayWithObject:@"download end\n"] 
				   name:NSStringFromSelector(_cmd)
			 resultsFor:target 
			   selector:sel];
	
	return [self __queueTransaction:[transaction autorelease] forClient:sender];
}

- restartDownload:sender
{
	//this method assumes that everything has been set up correctly. 
	//startDownloadingToPath ensures this.
	id		transaction;
	id		target = nil;
	SEL		sel = @selector(tomatoTorrentAgent:downloadStarted:);
	
	if ([sender respondsToSelector:sel])
		target = sender;
	
	transaction = [[TomatoTorrentAgentTransaction allocWithZone:[self zone]] 
		initWithStrings: [NSArray arrayWithObject:@"download begin\n"] 
				   name: NSStringFromSelector(_cmd)
			 resultsFor: target 
			   selector: sel];
	
	return [self __queueTransaction:[transaction autorelease] forClient:sender];
}

- downloadState:sender
{
	id		transaction;
	id		target = nil;
	SEL		sel = @selector(tomatoTorrentAgent:downloadState:);
	
	if ([sender respondsToSelector:sel])
		target = sender;
	
	transaction = [[TomatoTorrentAgentTransaction allocWithZone:[self zone]] 
		initWithStrings:[NSArray arrayWithObject:@"download state\n"] 
				   name:NSStringFromSelector(_cmd)
			 resultsFor:target 
			   selector:sel];
	
	return [self __queueTransaction:[transaction autorelease] forClient:sender];
}

- downloadConfigurationForKey:key delegate:sender;
{
	id		transaction;
	id		target = nil;
	SEL		sel = @selector(tomatoTorrentAgent:downloadConfiguration:);
	
	if ([sender respondsToSelector:sel])
		target = sender;

	transaction = [[TomatoTorrentAgentTransaction allocWithZone:[self zone]] 
		initWithStrings: [NSArray arrayWithObjects:[NSString stringWithFormat:@"download get %@\n", key], nil] 
				   name: NSStringFromSelector(_cmd)
			 resultsFor: target 
			   selector: sel];
	
	return [self __queueTransaction:[transaction autorelease] forClient:sender];
}

- configureDownloadKey:key withValue:value delegate:sender;
{
	id		transaction;
	id		target = nil;
	SEL		sel = @selector(tomatoTorrentAgent:configuredDownload:);
	
	if ([sender respondsToSelector:sel])
		target = sender;
	
	transaction = [[TomatoTorrentAgentTransaction allocWithZone:[self zone]] 
		initWithStrings: [NSArray arrayWithObject:[NSString stringWithFormat:@"download set %@ %@\n", key, value]] 
				   name: NSStringFromSelector(_cmd)
			 resultsFor: target 
			   selector: sel];
	
	return [self __queueTransaction:[transaction autorelease] forClient:sender];
}

// tracking
- startTracker:sender;
{
	id		transaction;
	id		target = nil;
	SEL		sel = @selector(tomatoTorrentAgent:trackerStartup:);
	
	if ([sender respondsToSelector:sel])
		target = sender;
	
	transaction = [[TomatoTorrentAgentTransaction allocWithZone:[self zone]] 
		initWithStrings:[self __beginTrackerStrings]
				   name:NSStringFromSelector(_cmd)
			 resultsFor:target 
			   selector:sel];
	
	__didStartTracker = YES;
	
	return [self __queueTransaction:[transaction autorelease] forClient:sender];
}

- stopTracker:sender;
{
	id		transaction;
	id		target = nil;
	SEL		sel = @selector(tomatoTorrentAgent:trackerStopped:);
	
	if ([sender respondsToSelector:sel])
		target = sender;
	
	transaction = [[TomatoTorrentAgentTransaction allocWithZone:[self zone]] 
		initWithStrings:[NSArray arrayWithObject:@"tracker stop\n"]
				   name:NSStringFromSelector(_cmd)
			 resultsFor:target 
			   selector:sel];
	
	return [self __queueTransaction:[transaction autorelease] forClient:sender];
}

- trackerState:sender;
{
	id		transaction;
	id		target = nil;
	SEL		sel = @selector(tomatoTorrentAgent:trackerState:);
	
	if ([sender respondsToSelector:sel])
		target = sender;
	
	transaction = [[TomatoTorrentAgentTransaction allocWithZone:[self zone]] 
		initWithStrings:[NSArray arrayWithObject:@"tracker state\n"]
				   name:NSStringFromSelector(_cmd)
			 resultsFor:target 
			   selector:sel];
	
	return [self __queueTransaction:[transaction autorelease] forClient:sender];
}

- (BOOL) didStartTracker;
{
	return __didStartTracker;
}

// generate
- startGenerate:sender;
{
	id		transaction;
	id		target = nil;
	SEL		sel = @selector(tomatoTorrentAgent:generateStartup:);
	
	if ([sender respondsToSelector:sel])
		target = sender;
	
	transaction = [[TomatoTorrentAgentTransaction allocWithZone:[self zone]] 
		initWithStrings:[self __beginGenerateStrings]
				   name:NSStringFromSelector(_cmd)
			 resultsFor:target 
			   selector:sel];
	
	return [self __queueTransaction:[transaction autorelease] forClient:sender];
}

- stopGenerate:sender;
{
	id		transaction;
	id		target = nil;
	SEL		sel = @selector(tomatoTorrentAgent:generateStopped:);
	
	if ([sender respondsToSelector:sel])
		target = sender;
	
	transaction = [[TomatoTorrentAgentTransaction allocWithZone:[self zone]] 
		initWithStrings:[NSArray arrayWithObject:@"generate stop\n"]
				   name:NSStringFromSelector(_cmd)
			 resultsFor:target 
			   selector:sel];
	
	return [self __queueTransaction:[transaction autorelease] forClient:sender];
}

- generateState:sender;
{
	id		transaction;
	id		target = nil;
	SEL		sel = @selector(tomatoTorrentAgent:generateState:);
	
	if ([sender respondsToSelector:sel])
		target = sender;
	
	transaction = [[TomatoTorrentAgentTransaction allocWithZone:[self zone]] 
		initWithStrings:[NSArray arrayWithObject:@"generate state\n"]
				   name:NSStringFromSelector(_cmd)
			 resultsFor:target 
			   selector:sel];
	
	return [self __queueTransaction:[transaction autorelease] forClient:sender];
}

- getInfoForMetaInfoFileAtPath:path delegate:sender;
{
	id		transaction;
	id		target = nil;
	SEL		sel = @selector(tomatoTorrentAgent:metainfo:);
	id		request;
	
	if ([sender respondsToSelector:sel])
		target = sender;
	
	request = [NSArray arrayWithObjects:
		[NSString stringWithFormat:@"set metainfo\n%@\n", path],
		@"metainfo\n", nil];
	
	transaction = [[TomatoTorrentAgentTransaction allocWithZone:[self zone]] 
		initWithStrings:request
				   name:NSStringFromSelector(_cmd)
			 resultsFor:target 
			   selector:sel];
	
	return [self __queueTransaction:[transaction autorelease] forClient:sender];
}

- (int)requestQueueSize;		//max number that can wait.
{
	return 7;
}

- (int)requestQueueBacklog
{
	int		result = -1;

	if ([__shutdownLock condition] == kShutdownLock_running)
	{
		[__transactionsLock lock];
			result = [__transactions count];
		[__transactionsLock unlock];
	}
	
	return result;
}

- (void)handleTakeValue:(id)value forUnboundKey:(NSString *)key;
{
	ttagent_logging_objc(0, @"TomatoTorrentAgent asked to take value for unknown key: %@", key);
}

@end


@implementation TomatoTorrentAgent (Private)

- __beginOperationStringsFromContainer:container name:optionCategory
{
	id		keys, objects, key, value, newCategory, result;
	
	result = [[NSMutableArray allocWithZone:[self zone]] initWithObjects:
				[NSString stringWithFormat:@"set %@\n\n", optionCategory], nil];	//clear existing.
	
	objects = [container objectEnumerator];
	if ([container respondsToSelector:@selector(keyEnumerator)])
		keys = [container keyEnumerator];
	else
		keys = [[[SKRange allocWithZone:[self zone]] initWithRange:NSMakeRange(0, [container count])]
					autorelease];

	while (value = [objects nextObject], key = [keys nextObject])
	{
		if ([value respondsToSelector:@selector(objectEnumerator)] == NO)
			[result addObject:[NSString stringWithFormat:@"set %@ %@\n%@\n", optionCategory, key, value]];
		else
		{
			newCategory = [[NSString alloc] initWithFormat:@"%@ %@", optionCategory, key];
			
				[result addObjectsFromArray:[self __beginOperationStringsFromContainer: value 
																				 name: newCategory]];
			[newCategory release];
		}
	}
	
	return [result autorelease];
}

- __beginDownloadStrings:destPath
{
	id		result, object;
	id		configWrapper;
	id		resumeDir;
	BOOL	isDir;
	
	configWrapper = [NSDictionary dictionaryWithObject: [self downloadConfigurationDictionary]
												forKey: @"configuration"];
	
	result = [self __beginOperationStringsFromContainer:configWrapper name:@"download"];
	
	//internal settings that are not managed through the dictionary
	object = [NSString stringWithFormat:@"set download configuration display_interval\n%f\n", __displayInterval];
	[result addObject:object];

	resumeDir = [g_torrent_data_dir stringByAppendingPathComponent:@"resume"];
	[[NSFileManager defaultManager] createDirectoryAtPath:resumeDir attributes:nil];
	
	if ([[NSFileManager defaultManager] fileExistsAtPath:resumeDir isDirectory:&isDir] && isDir)
	{
		object = [NSString stringWithFormat:@"set download configuration data_dir\n%@\n", g_torrent_data_dir];
		[result addObject:object];	
	}
	
	//+++ screwed if destPath contains a newline. urllib.unquote isn't smart enough to handle UTF8,
	//		which CFURLCreateStringByAddingPercentEscapes will escape.
	object = [NSString stringWithFormat:@"set download destPath\n%@\n", destPath];
	[result addObject:object];
	
	[result addObject:@"download begin\n"];
	
	return result;	
}

- __beginTrackerStrings
{
	id		result;

	NSAssert(__trackerDfile != nil, @"tried to start tracker without a state file!");
	
	result = [self __beginOperationStringsFromContainer:[self trackerOptions] name:@"tracker"];

	//internal settings that are not managed through the dictionary
	[result addObject:[NSString stringWithFormat:@"set tracker dfile\n%@\n", __trackerDfile]];
	if (__trackerAllowedDirPath)	[result addObject:[NSString stringWithFormat:@"set tracker allowed_dir\n%@\n", __trackerAllowedDirPath]];
	if (__trackerLogFilePath)	[result addObject:[NSString stringWithFormat:@"set tracker logfile\n%@\n", __trackerLogFilePath]];
	if (__trackerPort > 0)		[result addObject:[NSString stringWithFormat:@"set tracker port\n%d\n", __trackerPort]];

	[result addObject:@"tracker start\n"];
	return result;
}

- __beginGenerateStrings
{
	id		result;
	
	result = [self __beginOperationStringsFromContainer:[self generateOptions] name:@"generate"];

	[result addObject:[NSString stringWithFormat:@"set generate data_dir\n%@\n", g_torrent_data_dir]];

	[result addObject:@"generate start\n"];
	
	return result;
}

- __queueTransaction:transaction forClient:delegate
{
	id		result = nil;
	id		name = [transaction name];
	BOOL	needsToAdd = YES;
	int		numWaiting = 0;
	
	if ([__shutdownLock condition] == kShutdownLock_running)
	{
		[__transactionsLock lock];
		{	
			//coalesce these kinds of transactions. 
			if ([name isEqual:kDownloadStateTransactionName_] || [name isEqual:kDownloadConfigTransactionName_])
			{
				unsigned index = [__transactions indexOfObject:transaction];
				
				if (index != NSNotFound)
				{
					[[__transactions objectAtIndex:index] incrementCount];

					needsToAdd = NO;
					result = self;
				}
			}
			
			if (needsToAdd && [__transactions count] <= 7)
			{
				[__transactions insertObject:transaction atIndex:0];
				result = self;
			}
			else if (__numEffectiveTransactions > [__transactions count])
			{
				//backlogged
				numWaiting = __numEffectiveTransactions;
			}
			
			if (result != nil)
				++__numEffectiveTransactions;			
		}
		[__transactionsLock unlockWithCondition:kQueueHasTransactions];
		
		if (numWaiting)
		{
			SEL		foo = @selector(tomatoTorrentAgent:backloggedRequest:wasRejected:backlogCount:);

			if (delegate != nil && [delegate respondsToSelector:foo])
			{
				[delegate tomatoTorrentAgent: self
						backloggedRequest: [transaction name]
							  wasRejected: result == nil
							 backlogCount: numWaiting];
			}
			
			[[NSNotificationCenter defaultCenter] postNotificationName:[transaction name] object:self];
		}
	}
	return result;
}

- __startTaskThread
{
	__transactionsLock = [[NSConditionLock allocWithZone:[self zone]] init];
	__transactions = [[NSMutableArray allocWithZone:[self zone]] init];

	__mainThreadPort = [[NSPort allocWithZone:[self zone]] init];

	[__mainThreadPort setDelegate:self];
	[[NSRunLoop currentRunLoop] addPort:__mainThreadPort forMode:NSDefaultRunLoopMode];

	[NSThread detachNewThreadSelector:@selector(__taskThread:) toTarget:self withObject:nil];
	return self;
}

- __releasePortMessageObjects:message
{
	id		target, directParam, param;
	
	if (message)
	{
		[[[message components] objectAtIndex:0] getBytes:&target];
		[[[message components] objectAtIndex:2] getBytes:&directParam];
		[[[message components] objectAtIndex:3] getBytes:&param];	
		
		[target release];
		[directParam release];
		[param release];
	}
	
	return self;
}

- (void)handlePortMessage:(NSPortMessage *)message;
{
	id		target, directParam, param;
	SEL		selector;

	[[[message components] objectAtIndex:0] getBytes:&target];
	[[[message components] objectAtIndex:1] getBytes:&selector];
	[[[message components] objectAtIndex:2] getBytes:&directParam];
	[[[message components] objectAtIndex:3] getBytes:&param];
	
	[target performSelector:selector withObject:directParam withObject:param];
	
	//[self __releasePortMessageObjects:message];
	[target release];
	[directParam release];
	[param release];
}

- __stopTaskThread
{
	//remove all transactions and set the condition to non-empty. that tells it to stop.
	if ([__shutdownLock condition] == kShutdownLock_running)
	{
		[__transactionsLock lock];
			[__transactions removeAllObjects];
		[__transactionsLock unlockWithCondition:kQueueHasTransactions];
	}
	
	[[NSNotificationCenter defaultCenter] removeObserver:self];
	
	//wait a limited time for it to shutdown properly.
	if ([__shutdownLock lockWhenCondition:kShutdownLock_shutdown
						  beforeDate:[NSDate dateWithTimeIntervalSinceNow:10]])
	{
		[__shutdownLock unlock];
	}
	else
	{
		NSLog(@"agent thread not shutting down after 10 seconds!");
	}
	
	//thread releases other locks when it wants to.
	return self;
}

//------------------------------------------------------------------------------
//		• __taskThread:
//------------------------------------------------------------------------------
//	Move along, nothing to see here. Check out __taskLoop.

- (void) __taskThread:param
{
	NSAutoreleasePool	*pool;
	SEL					post = @selector(postResponse:forAgent:);
	
	pool = [[NSAutoreleasePool allocWithZone:[self zone]] init];
	{
		ttagent_logging_objc(1, @"%x: starting task loop.", self);
	
		[[self retain] autorelease];	//to properly dealloc, you have to send a shutdown message to the object.
		[__transactions autorelease];
		[__transactionsLock autorelease];
		
		__postSelectorData = [NSData dataWithBytes:&post length:sizeof(post)];
		__agentData = [NSData dataWithBytes:&self length:sizeof(self)];
	
		[self __taskLoop];
		
		//clean up stuff.
		__transactions = __transactionsLock = nil;
		
		//sync shutdown without join (allow failure).
		[__shutdownLock lock];
		[__shutdownLock unlockWithCondition:kShutdownLock_shutdown];
		
		[__mainThreadPort invalidate], [__mainThreadPort release];
	}
	[pool release];
	
	ttagent_logging_objc(1, @" thread code ends here ");
	ttagent_logging_objc(1, @"----------snip---------");

	[NSThread exit];
}

//------------------------------------------------------------------------------
//		• __taskLoop
//------------------------------------------------------------------------------
//	Move along, nothing to see here. See __taskGatewayLoop.

- __taskLoop
{
	char*	error 			= NULL;

   __agent = ttagent_startup([TTAgentPythonLibPath([self class]) UTF8String], &error);

    if (__agent)
	{
		[self __taskGatewayLoop];
		ttagent_shutdown(__agent);
	}
	else
	{
		NSLog(@"BitTorrent task did not run; %s", error);
		free(error);
	}
	
	return self;
}

//------------------------------------------------------------------------------
//		• __taskGatewayLoop
//------------------------------------------------------------------------------
//	Loops while kQueueHasTransactions AND __transactions is not empty.
//
//		During each iteration of the loop, it either fetches a transaction
//		and creates a message, or sends a current message. This is because
//		a message may fail to send right away. Rather than waiting forever
//		for the message to send, it keeps looping and sees if it should
//		terminate the loop.

- __taskGatewayLoop
{
	id		transaction = nil, 
			message = nil, 
			pool;
	int		numWaiting;
	BOOL	done = NO;
	NSZone	*zone = [self zone];

	
	do
	{
		pool = [[NSAutoreleasePool allocWithZone:zone] init];
		
		ttagent_logging_objc(10, @"waiting for transaction");
		
		[__transactionsLock lockWhenCondition:kQueueHasTransactions];
		{
			numWaiting = [__transactions count];
			if (numWaiting > 0)
			{
				if (message == nil)
				{
					transaction = [[__transactions lastObject] retain];
					[__transactions removeLastObject];
					--numWaiting;
					
					__numEffectiveTransactions -= [transaction effectiveCount];
				}
				//else
					//we're free to continue trying to send the message.
			}
			else
			{
				//else if 'got go' when the queue actually was empty, we have to exit the loop.
				[self __releasePortMessageObjects:message];
				[message release];
				message = nil;
				done = YES;
			}
		}	
		[__transactionsLock unlockWithCondition: 
			(numWaiting > 0 ? kQueueHasTransactions : kTransactionQueueEmpty)];
		
		ttagent_logging_objc(5, @"%x: transaction %x (%d waiting), message %x", 
			self, transaction, numWaiting, message);
		
		if (message == nil && transaction != nil)
		{
			//process transaction
			message = [self __processTransaction:transaction];

			[transaction release];
			transaction = nil;
		}
		
		if (message)
		{
			if ([message sendBeforeDate:[NSDate dateWithTimeIntervalSinceNow:7]])
			{
				[message release];
				message = nil;
			}
		}
		else
			done = YES;
			//terminate loop if transaction processing results in a hard failure (no message).
		
		[pool release];
	} 
	while (!done);


	return self;
}

//------------------------------------------------------------------------------
//		• __processTransaction
//------------------------------------------------------------------------------
//	On any fault, it makes sure that the response object includes an entry for 'fault',
//	and on hard failures (task is returning super-funky data) it returns nil, indicating
//	that processing should stop because the system's unreliable. 
//
//	On success, returns the message response to send.
//
//		I was using NSData/NSString/NSFileHandle, but the code was HUGE and the unix stuff
//		is much better. Besides, Cocoa only exists on POSIX-compatible platforms right now :P

- __processTransaction:transaction
{
	id			string, ite, response, result;		// result is not a return value. 
	BOOL		done = NO, 							// It's used for a lot of things. 
				hosed = NO;							// Too many things.
	NSZone		*zone = [self zone];
	
	id			message = nil;		//return value.
	id			startDate;


	startDate = [NSDate date];

	//send each item in strings, and if a fault's found in the response we abort the whole batch.
	//we send the response consisting of the last result we got.
	//
	//	if a fault occured while actually trying to get a response, we terminate the thread loop.
	
	ite = [[transaction strings] objectEnumerator];

	result = nil;
	string = [ite nextObject];
	if (string == nil) done = YES;	//wtf?
	
	while (!done)
	{	
		ttagent_logging_objc(5, @"%x: sending request %@", self, string);
		
		if (ttagent_send_message(__agent, [string UTF8String]))
		{
			char	*answer;
			
			ttagent_logging_objc(10, @"%x: waiting for answer", self);
			
			answer = ttagent_read_answer(__agent, NULL);
			
			ttagent_logging_objc(10, @"%x: got answer, len: %d", self, strlen(answer));
			
			if (answer)
			{
				id 	data, errorString = nil;
				
				data = [[NSData allocWithZone:zone] initWithBytes:answer length:strlen(answer)];
				
				result = (id)CFPropertyListCreateFromXMLData(
					NULL, (CFDataRef)data, kCFPropertyListImmutable, (CFStringRef*)&errorString);

				if (result)
				{
					[result autorelease];				
					//check result for faults, in which case we're done with this transaction
					if ([result respondsToSelector:@selector(objectForKey:)])
					{
						if ([result objectForKey:@"fault"] != nil)
							done = YES;
					}
					else
					{
						result = [NSDictionary dictionaryWithObject:[NSArray arrayWithObjects:
										@"Got a plist from the task, but it doesn't seem to be a dictionary.", result, nil]
									forKey:@"fault"];
						done = YES;	//not a hard failure, since we did seem to get a plist. Which is valid.
					}
				}
				else
				{
					id	string = [[[NSString alloc] initWithData:data encoding:NSUTF8StringEncoding] autorelease];
					//[data writeToFile:@"/tmp/badplist.text" atomically:NO];
					
					result = [NSDictionary dictionaryWithObject:[NSArray arrayWithObjects:
									@"Couldn't make plist from task output. Hosed.",
									(errorString ? errorString : @"no errorString"), string, nil]
								forKey:@"fault"];
					[errorString release];
					done = hosed = YES;
				}
				
				[data release];
				free(answer);
			}
			else
				done = hosed = YES, result = nil;
		}
		else
			done = hosed = YES, result = nil;

		string = [ite nextObject];
		done = (string == nil);
	}

	if (hosed && result == nil && __agent->last_error_desc)
	{
		result = [NSDictionary dictionaryWithObject:[NSArray arrayWithObject:[NSString stringWithCString:__agent->last_error_desc]]
											 forKey:@"fault"];
	}
	else if (result == nil) //wtf
	{
		result = [NSDictionary dictionaryWithObject:@"Goofed up; no response data."
											 forKey:@"fault"];
	}
	
	response = [[TomatoTorrentAgentResponse alloc] initWithResponseDictionary:result 
															  respondingTo:[transaction name]];

	[response setConversationDuration:[startDate timeIntervalSinceNow]];
	
	//[transaction postResponse:response forAgent:self];	//not used because of 10.1 compat.
															//it used to use performSelectorInMainThread.
	message = nil;
	
	if (!hosed)
	{
		id		transData, respData, components; 
		
		transData 	= [NSData dataWithBytes:&transaction length:sizeof(transaction)];
		respData 	= [NSData dataWithBytes:&response length:sizeof(response)];
		components 	= [NSArray arrayWithObjects:transData, __postSelectorData, respData, __agentData, nil];

		message = [[NSPortMessage alloc] initWithSendPort:__mainThreadPort receivePort:nil components:components];
		
		//released on receive.
		[response retain];
		[transaction retain];
		[self retain];
	}
	
	ttagent_logging_objc(kTomatoAgentLoggingLevel_Data, @"%x: response: %@", self, response);
	[response release];

	return message;
	
}

+ __processDefaultsAnswer:answer
{
	id		result;
	id		exceptions	= [answer objectForKey:kDefaultsAnswerKey_exceptions],
			defaults	= [answer objectForKey:kDefaultsAnswerKey_defaults],
			bools 		= [answer objectForKey:kDefaultsAnswerKey_boolKeys];
	id		ite, object, list, map;

	map = [NSMutableDictionary dictionary];
	list = [NSMutableArray array];
	ite = [defaults objectEnumerator];
	
	while (object = [ite nextObject])
	{
		id		key = [object objectAtIndex:0];
		
		if ([exceptions containsObject:key] == NO)
		{
			[list addObject:object];
			[map setObject:object forKey:key];
		}
	}
	
	//immutable copies
	map = [map copy];
	list = [list copy];
	bools = [bools copy];
	
	//now put them in a dictionary.
	result = [[NSDictionary alloc] initWithObjectsAndKeys:
		list,		kTomatoAgentModuleDefaultsKey_list,
		map,		kTomatoAgentModuleDefaultsKey_map, 
		bools,		kTomatoAgentModuleDefaultsKey_boolKeys,
		nil];
	
	[map release];
	[list release];
	[bools release];
	
	return [result autorelease];
}

//------------------------------------------------------------------------------
//		• __runTaskToFetchDefaults:
//------------------------------------------------------------------------------
//	this is a thread.

+ (void) __runTaskToFetchDefaults:params
{
	char				*error = NULL;
	char				*answer = NULL;
	ttagent_context		*defaultsAgent;
	id					pool, 
						mutex, 
						defaultsCommand,
						result = nil;
	
	pool = [NSAutoreleasePool new];
	
	[[params retain] autorelease];

	mutex = [params objectForKey:@"mutex"];
	defaultsCommand = [params objectForKey:@"command"];
	
	[mutex lock];
		defaultsAgent = ttagent_startup([TTAgentPythonLibPath([self class]) UTF8String], &error);
		if (defaultsAgent)
		{
			ttagent_send_message(defaultsAgent, [defaultsCommand UTF8String]);
			answer = ttagent_read_answer(defaultsAgent, NULL);
			if (answer)
			{
				id		data = [NSData dataWithBytes:answer length:strlen(answer)];
				id		errorString = NULL;
				
				result = (id)CFPropertyListCreateFromXMLData(
					NULL, (CFDataRef)data, kCFPropertyListImmutable, (CFStringRef*)&errorString);
				
				if (errorString)
				{
					NSLog(@"couldn't read defaults plist: %@", errorString);
					[errorString autorelease];
				}
			}
			
			ttagent_shutdown(defaultsAgent);
		}
		else
		{
			NSLog(@"BitTorrent task did not run; %s", error);
			free(error);
		}
		
		if (result)
		{
			//turn it into a response then process the data.
			result = [[TomatoTorrentAgentResponse alloc] initWithResponseDictionary:[result autorelease]
																	respondingTo:nil];
							  
			[params setObject:[self __processDefaultsAnswer:[result responseContent]]
					   forKey:@"defaults"];
			
			[result release];
		}
	[mutex unlockWithCondition:1];
	
	[pool release];
}


@end

@implementation TomatoTorrentAgentTransaction

- initWithStrings:data name:name resultsFor:target selector:(SEL)sel;
{
	[super init];

	__data = [data retain];
	__name = [name retain];
	__target = [target retain];
	__action = sel;
	__count = 1;
	
	return self;
}

- (void) dealloc
{
	[__target release];
	[__name release];
	[__data release];
	[super dealloc];
}


- strings;
{
	return __data;
}

- name
{
	return __name;
}

- (unsigned) effectiveCount
{
	return __count;
}

- incrementCount
{
	++__count;
	return self;
}

- (BOOL) isEqual:transaction
{
	BOOL	result = NO;
	
	if ([transaction isKindOfClass:[self class]])
	{
		TomatoTorrentAgentTransaction*	foo = (TomatoTorrentAgentTransaction*)transaction;
		
		result = [foo->__name isEqual:__name] && foo->__target == __target && foo->__action == __action;
	}
	
	return result;	
}

- postResponse:response forAgent:agent
{
	if (__target)
	{
		[__target performSelector:__action withObject:agent withObject:response];
	}
	
	if (__name)
	{
		[[NSNotificationCenter defaultCenter] postNotificationName:(id)[response request]
															object:self
														  userInfo:[NSDictionary dictionaryWithObject:response
																							   forKey:kTomatoAgentNotifKey_response]];	
	}
	
	#if defined(__TomatoDebug_)
	if ([response fault])
		ttagent_logging_objc(5, @"%x: response fault: %@", agent, response);
	#endif
	
	return self;
}

@end

@implementation TomatoTorrentAgentResponse
- initWithResponseDictionary:data respondingTo:name;
{
	[super init];
	__response = [data copyWithZone:[self zone]];
	__name = [name copyWithZone:[self zone]];
	return self;
}

- copyWithZone:(NSZone*)zone
{
	id		result;
	
	result = [[TomatoTorrentAgentResponse allocWithZone:zone] 
		initWithResponseDictionary:__response respondingTo:__name];
	
	return result;
}

- (void) dealloc
{
	[__response release];
	[__name release];
	[super dealloc];
}

- responseContent;			//'data' key of response
{
	return [__response objectForKey:@"data"];
}

- fault;		//'fault' key
{
	return [__response objectForKey:@"fault"];
}

- request;		//name of selector that caused the response
{
	return __name;
}

- responseDictionary;
{
	return __response;
}

- description
{
	return [NSString stringWithFormat:@"<TomatoTorrentAgentResponse:%x %@>", self, __response];
}

- setConversationDuration:(double)duration;
{
	__duration = duration;
	return self;
}

- (double) conversationDuration
{
	return __duration;
}

@end


static id TTAgentPythonLibPath(Class agent)
{
	return [[NSBundle bundleForClass:agent] resourcePath];
}

#if defined(__SK_genstrings_cpp__)
	
	//\#define.(k[^\t]*)[^@]+(@\"[^\"]+\").*$

void
tomatoTorrentAgent_localized_strings()
{
	TTAgentLocalizedString_(kTomatoAgentKey_downloadState_name, kTomatoAgentStringsTable_downloadState);
	TTAgentLocalizedString_(kTomatoAgentKey_downloadState_size, kTomatoAgentStringsTable_downloadState);
	TTAgentLocalizedString_(kTomatoAgentKey_downloadState_saveas, kTomatoAgentStringsTable_downloadState);
	TTAgentLocalizedString_(kTomatoAgentKey_downloadState_isDirectory, kTomatoAgentStringsTable_downloadState);
	
	//rest of the keys that I know about
	TTAgentLocalizedString_(kTomatoAgentKey_downloadState_errors, kTomatoAgentStringsTable_downloadState);
	TTAgentLocalizedString_(kTomatoAgentKey_downloadState_fault, kTomatoAgentStringsTable_downloadState);
	TTAgentLocalizedString_(kTomatoAgentKey_downloadState_destPath, kTomatoAgentStringsTable_downloadState);
	TTAgentLocalizedString_(kTomatoAgentKey_downloadState_status, kTomatoAgentStringsTable_downloadState);
	TTAgentLocalizedString_(kTomatoAgentKey_downloadState_statusSetAtTime, kTomatoAgentStringsTable_downloadState);
	TTAgentLocalizedString_(kTomatoAgentKey_downloadState_status_starting, kTomatoAgentStringsTable_downloadState);
	TTAgentLocalizedString_(kTomatoAgentKey_downloadState_status_connecting, kTomatoAgentStringsTable_downloadState);
	TTAgentLocalizedString_(kTomatoAgentKey_downloadState_status_downloading, kTomatoAgentStringsTable_downloadState);
	TTAgentLocalizedString_(kTomatoAgentKey_downloadState_status_finished, kTomatoAgentStringsTable_downloadState);
	TTAgentLocalizedString_(kTomatoAgentKey_downloadState_status_failed, kTomatoAgentStringsTable_downloadState);
	TTAgentLocalizedString_(kTomatoAgentKey_downloadState_status_limbo, kTomatoAgentStringsTable_downloadState);
	TTAgentLocalizedString_(kTomatoAgentKey_downloadState_status_halted, kTomatoAgentStringsTable_downloadState);
	
	TTAgentLocalizedString_(kTomatoAgentKey_downloadState_stats, kTomatoAgentStringsTable_downloadState);
	TTAgentLocalizedString_(kTomatoAgentKey_downloadState_stats_fractionDone, kTomatoAgentStringsTable_downloadState);
	TTAgentLocalizedString_(kTomatoAgentKey_downloadState_stats_activity, kTomatoAgentStringsTable_downloadState);
	TTAgentLocalizedString_(kTomatoAgentKey_downloadState_stats_timeEstimate, kTomatoAgentStringsTable_downloadState);
	TTAgentLocalizedString_(kTomatoAgentKey_downloadState_stats_downRate, kTomatoAgentStringsTable_downloadState);
	TTAgentLocalizedString_(kTomatoAgentKey_downloadState_stats_upRate, kTomatoAgentStringsTable_downloadState);
	TTAgentLocalizedString_(kTomatoAgentKey_downloadState_stats_downTotal, kTomatoAgentStringsTable_downloadState);

	TTAgentLocalizedString_(kTomatoAgentKey_downloadState_stats_spew, kTomatoAgentStringsTable_downloadState);
	TTAgentLocalizedString_(kTomatoAgentKey_downloadState_stats_spew_download, kTomatoAgentStringsTable_downloadState);
	TTAgentLocalizedString_(kTomatoAgentKey_downloadState_stats_spew_unchoke, kTomatoAgentStringsTable_downloadState);
	TTAgentLocalizedString_(kTomatoAgentKey_downloadState_stats_spew_ip, kTomatoAgentStringsTable_downloadState);
	TTAgentLocalizedString_(kTomatoAgentKey_downloadState_stats_spew_initiation, kTomatoAgentStringsTable_downloadState);
	TTAgentLocalizedString_(kTomatoAgentKey_downloadState_stats_spew_upload, kTomatoAgentStringsTable_downloadState);

	TTAgentLocalizedString_(kTomatoAgentKey_downloadState_stats_spew_initiation_remote, kTomatoAgentStringsTable_downloadState);
	TTAgentLocalizedString_(kTomatoAgentKey_downloadState_stats_spew_initiation_local, kTomatoAgentStringsTable_downloadState);
	
	TTAgentLocalizedString_(kTomatoAgentKey_tracker_dfile, kTomatoAgentStringsTable_tracker);
	TTAgentLocalizedString_(kTomatoAgentKey_tracker_port, kTomatoAgentStringsTable_tracker);
	TTAgentLocalizedString_(kTomatoAgentKey_tracker_allowedDir, kTomatoAgentStringsTable_tracker);
	TTAgentLocalizedString_(kTomatoAgentKey_tracker_logfile, kTomatoAgentStringsTable_tracker);
	
	TTAgentLocalizedString_(kTomatoAgentKey_trackerState_status, kTomatoAgentStringsTable_tracker);
	TTAgentLocalizedString_(kTomatoAgentKey_trackerState_statusSetAtTime, kTomatoAgentStringsTable_tracker);
	
	TTAgentLocalizedString_(kTomatoAgentKey_trackerState_status_running, kTomatoAgentStringsTable_tracker);
	TTAgentLocalizedString_(kTomatoAgentKey_trackerState_status_limbo, kTomatoAgentStringsTable_tracker);
	TTAgentLocalizedString_(kTomatoAgentKey_trackerState_status_none, kTomatoAgentStringsTable_tracker);
	
	TTAgentLocalizedString_(kTomatoAgentKey_generate_files, kTomatoAgentStringsTable_generate);
	TTAgentLocalizedString_(kTomatoAgentKey_generate_filesAttribs, kTomatoAgentStringsTable_generate);
	TTAgentLocalizedString_(kTomatoAgentKey_generate_destDirPath, kTomatoAgentStringsTable_generate);
	TTAgentLocalizedString_(kTomatoAgentKey_generate_trackerURL, kTomatoAgentStringsTable_generate);
	TTAgentLocalizedString_(kTomatoAgentKey_generate_pieceSizePow2, kTomatoAgentStringsTable_generate);
	
	TTAgentLocalizedString_(kTomatoAgentKey_generate_fileAttrib_destDirPath, kTomatoAgentStringsTable_generate);
	TTAgentLocalizedString_(kTomatoAgentKey_generate_fileAttrib_trackerURL, kTomatoAgentStringsTable_generate);
	TTAgentLocalizedString_(kTomatoAgentKey_generate_fileAttrib_comment, kTomatoAgentStringsTable_generate);
	
	TTAgentLocalizedString_(kTomatoAgentKey_generateState_status, kTomatoAgentStringsTable_generate);
	TTAgentLocalizedString_(kTomatoAgentKey_generateState_statusSetAtTime, kTomatoAgentStringsTable_generate);
	
	TTAgentLocalizedString_(kTomatoAgentKey_generateState_status_limbo, kTomatoAgentStringsTable_generate);
	TTAgentLocalizedString_(kTomatoAgentKey_generateState_status_finished, kTomatoAgentStringsTable_generate);
	TTAgentLocalizedString_(kTomatoAgentKey_generateState_status_running, kTomatoAgentStringsTable_generate);
	TTAgentLocalizedString_(kTomatoAgentKey_generateState_status_failed, kTomatoAgentStringsTable_generate);
	
		//if you have one of these in the response you have them all.
	TTAgentLocalizedString_(kTomatoAgentKey_generateState_numFiles, kTomatoAgentStringsTable_generate);
	TTAgentLocalizedString_(kTomatoAgentKey_generateState_totalBytes, kTomatoAgentStringsTable_generate);
	TTAgentLocalizedString_(kTomatoAgentKey_generateState_completedBytes, kTomatoAgentStringsTable_generate);
	TTAgentLocalizedString_(kTomatoAgentKey_generateState_currentItemPath, kTomatoAgentStringsTable_generate);
	TTAgentLocalizedString_(kTomatoAgentKey_generateState_currentItemNum, kTomatoAgentStringsTable_generate);
	
	TTAgentLocalizedString_(kTomatoAgentKey_generateState_errors, kTomatoAgentStringsTable_generate);
	
	TTAgentLocalizedString_(kTomatoAgentKey_metainfo_arg, kTomatoAgentStringsTable_metainfo);
	
	TTAgentLocalizedString_(kTomatoAgentKey_metainfo_arg_announce, kTomatoAgentStringsTable_metainfo);
	TTAgentLocalizedString_(kTomatoAgentKey_metainfo_arg_comment, kTomatoAgentStringsTable_metainfo);
	TTAgentLocalizedString_(kTomatoAgentKey_metainfo_arg_creationDate, kTomatoAgentStringsTable_metainfo);
	TTAgentLocalizedString_(kTomatoAgentKey_metainfo_info, kTomatoAgentStringsTable_metainfo);
	TTAgentLocalizedString_(kTomatoAgentKey_metainfo_isDirectory, kTomatoAgentStringsTable_metainfo);
	TTAgentLocalizedString_(kTomatoAgentKey_metainfo_arg_totalBytes, kTomatoAgentStringsTable_metainfo);
	TTAgentLocalizedString_(kTomatoAgentKey_metainfo_numPieces, kTomatoAgentStringsTable_metainfo);
	TTAgentLocalizedString_(kTomatoAgentKey_metainfo_lastPieceLength, kTomatoAgentStringsTable_metainfo);
	
	//metainfo info keys
	TTAgentLocalizedString_(kTomatoAgentKey_metainfo_info_pieceLength, kTomatoAgentStringsTable_metainfo);
	TTAgentLocalizedString_(kTomatoAgentKey_metainfo_info_name, kTomatoAgentStringsTable_metainfo);
	TTAgentLocalizedString_(kTomatoAgentKey_metainfo_info_pieces, kTomatoAgentStringsTable_metainfo);
	TTAgentLocalizedString_(kTomatoAgentKey_metainfo_info_length, kTomatoAgentStringsTable_metainfo);
	TTAgentLocalizedString_(kTomatoAgentKey_metainfo_info_files, kTomatoAgentStringsTable_metainfo);
	TTAgentLocalizedString_(kTomatoAgentKey_metainfo_info_files_length, kTomatoAgentStringsTable_metainfo);
	TTAgentLocalizedString_(kTomatoAgentKey_metainfo_info_files_path, kTomatoAgentStringsTable_metainfo);

	//exceptions.
	SKLocalizedStringWithDefaultValue(@"unknown url type: ", kTomatoAgentStringsTable_responseFaults, TTAgentBundle_(), 
		@"unknown url type: %@", @"unknown url exception (default: unknown url type: %@)");

	//typos, in case if they get fixed
	TTAgentLocalizedString_(@"initiation", kTomatoAgentStringsTable_downloadState);

}
#endif
